# -*- coding: utf-8 -*-
"""Utility functions."""

import re
import random
import numpy
import chardet
from math import log
from chardet.universaldetector import UniversalDetector
from numpy import fft
from scipy.signal import find_peaks
from .exceptions import ConsistencyException, FloatConversionError
from datetime import datetime
from propertime.utils import s_from_dt
import subprocess
from collections import namedtuple
import numpy as np
from scipy import stats, optimize
import scipy.stats
import matplotlib.pyplot as plt
from sklearn.metrics import mean_squared_error as sklearn_mean_squared_error
from sklearn.metrics import mean_absolute_error as sklearn_mean_absolute_error


# Setup logging
import logging
logger = logging.getLogger(__name__)


def ensure_reproducibility():
    """Ensure reproducibility by fixing seeds and initial conditions for: Random, Numpy, Keras and Tensorflow."""

    random.seed(0)
    numpy.random.seed(0)
    try:
        import tensorflow
    except ImportError:
        pass
    else:
        # Ensure reproducibility for Tensorflow as well
        tensorflow.random.set_seed(0)


def is_numerical(item):
    """Check if item is numerical (float or int, including Pandas data types)."""
    if isinstance(item, float):
        return True
    if isinstance(item, int):
        return True
    try:
        # Handle pandas-like data types (i.e. int64)
        item + 1
        return True
    except:
        return False


def detect_encoding(file_name, streaming=False):
    """Detect the encoding of a file."""
    if streaming:
        detector = UniversalDetector()
        with open(file_name, 'rb') as file_pointer:
            for i, line in enumerate(file_pointer.readlines()):
                logger.debug('Itearation #%s: confidence=%s',i,detector.result['confidence'])
                detector.feed(line)
                if detector.done:
                    logger.debug('Detected encoding at line "%s"', i)
                    break
        detector.close()
        chardet_results = detector.result

    else:
        with open(file_name, 'rb') as file_pointer:
            chardet_results = chardet.detect(file_pointer.read())

    logger.debug('Detected encoding "%s" with "%s" confidence (streaming=%s)', chardet_results['encoding'],chardet_results['confidence'], streaming)
    encoding = chardet_results['encoding']

    return encoding


def detect_sampling_interval(timeseries, confidence=False):
    """Detect the sampling interval of a time series."""

    diffs={}
    prev_point=None
    for point in timeseries:
        if prev_point is not None:
            diff = point.t - prev_point.t
            if diff not in diffs:
                diffs[diff] = 1
            else:
                diffs[diff] +=1
        prev_point = point

    # Iterate until the diffs are not too spread, then pick the maximum.
    i=0
    while _is_almost_equal(len(diffs), len(timeseries)):
        or_diffs=diffs
        diffs={}
        for diff in or_diffs:
            diff=round(diff)
            if diff not in diffs:
                diffs[diff] = 1
            else:
                diffs[diff] +=1

        if i > 10:
            raise Exception('Cannot automatically detect the sampling interval')

    most_common_diff_total = 0
    most_common_diff = None
    for diff in diffs:
        if diffs[diff] > most_common_diff_total:
            most_common_diff_total = diffs[diff]
            most_common_diff = diff

    second_most_common_diff_total = 0
    second_most_common_diff = None
    for diff in diffs:
        if diff == most_common_diff:
            continue
        if diffs[diff] > second_most_common_diff_total:
            second_most_common_diff_total = diffs[diff]
            second_most_common_diff = diff

    if confidence:
        if len(diffs) == 1:
            confidence_value = 1.0
        else:
            # TODO: this is a totally arbitrary confidence metric, check me.
            confidence_value = 1 - (second_most_common_diff_total / most_common_diff_total)
        return most_common_diff, confidence_value
    else:
        return most_common_diff


def detect_periodicity(timeseries):
    """Detect the periodicity of a time series."""

    _check_timeseries(timeseries)

    # TODO: fix me, data_loss must not belong as data_label
    data_labels = timeseries.data_labels()

    if len(data_labels) > 1:
        raise NotImplementedError()

    # TODO: improve me, highly ineficcient
    for data_label in data_labels:

        # Get data as a vector
        y = []
        for item in timeseries:
            y.append(item.data[data_label])
        #y = [item.data[data_label] for item in timeseries]

        # Compute FFT (Fast Fourier Transform)
        yf = fft.fft(y)

        # Remove specular data
        len_yf = len(yf)
        middle_point=round(len_yf/2)
        yf = yf[0:middle_point]

        # To absolute values
        yf = [abs(f) for f in yf]

        # Find FFT peaks
        peak_indexes, _ = find_peaks(yf, height=None)
        peaks = []
        for i in peak_indexes:
            peaks.append([i, yf[i]])

        # Sort by peaks intensity and compute actual frequency in base units
        # TODO: round peak frequencies to integers and/or neighbours first
        peaks = sorted(peaks, key=lambda t: t[1])
        peaks.reverse()

        # Compute peak frequencies:
        for i in range(len(peaks)):

            # Set peak frequency
            peak_frequency = (len(y) / peaks[i][0])
            peaks[i].append(peak_frequency)

        # Find most relevant frequency
        max_peak_frequency = None
        for i in range(len(peaks)):

            logger.debug('Peak #%s: \t index=%s,\t value=%s, freq=%s (over %s)', i, peaks[i][0], int(peaks[i][1]), peaks[i][2], len(timeseries))

            # Do not consider lower frequencies if there is a closer and higher one
            try:
                diff1=peaks[i][1]-peaks[i+1][1]
                diff2=peaks[i+1][1]-peaks[i+2][1]
                if diff1 *3 < diff2:
                    logger.debug('Peak #{} candidate to removal'.format(i))
                    if (peaks[i][2] > peaks[i+1][2]*10) and (peaks[i][2] > len(timeseries)/10):
                        logger.debug('peak #{} marked to be removed'.format(i))
                        continue
            except IndexError:
                pass

            if not max_peak_frequency:
                max_peak_frequency = peaks[i][2]
            if i>10:
                break

        if not max_peak_frequency:
            raise ValueError('Cannot detect any periodicity!')

        # Round max peak and return
        return int(round(max_peak_frequency))


def mean_absolute_percentage_error(list1, list2):
    """Compute the MAPE, list 1 are true values, list 2 are predicted values."""

    if len(list1) != len(list2):
        raise ValueError('Lists have different lengths, cannot continue')
    p_error_sum = 0
    for i in range(len(list1)):
        p_error_sum += abs((list1[i] - list2[i])/list1[i])
    return p_error_sum/len(list1)


def max_absolute_percentage_error(list1, list2):
    """Compute the MaxAPE, list 1 are true values, list 2 are predicted values."""

    if len(list1) != len(list2):
        raise ValueError('Lists have different lengths, cannot continue')
    max_ape = None
    for i in range(len(list1)):
        ape = abs((list1[i] - list2[i])/list1[i])
        if max_ape is None:
            max_ape = ape
        else:
            if ape > max_ape:
                max_ape = ape
    return max_ape


def mean_absolute_error(list1,list2):
    """Compute the MAE, list 1 are true values, list 2 are predicted values."""
    return sklearn_mean_absolute_error(list1,list2)


def max_absolute_error(list1,list2):
    """Compute the MaxAE, list 1 are true values, list 2 are predicted values."""
    if len(list1) != len(list2):
        raise ValueError('Lists have different lengths, cannot continue')
    max_ae = None
    for i in range(len(list1)):
        ae = abs(list1[i] - list2[i])
        if max_ae is None:
            max_ae = ae
        else:
            if ae > max_ae:
                max_ae = ae
    return max_ae


def mean_absolute_log_error(list1, list2):
    """Compute the MALE, list 1 are true values, list 2 are predicted values."""
    if len(list1) != len(list2):
        raise ValueError('Lists have different lengths, cannot continue')
    ale_sum = 0
    for i in range(len(list1)):
        ale_sum += abs(log(list1[i]/list2[i]))
    return ale_sum/len(list1)


def max_absolute_log_error(list1,list2):
    """Compute the MaxALE, list 1 are true values, list 2 are predicted values."""
    if len(list1) != len(list2):
        raise ValueError('Lists have different lengths, cannot continue')
    max_ale = None
    for i in range(len(list1)):
        ale = abs(log(list1[i]/list2[i]))
        if max_ale is None:
            max_ale = ale
        else:
            if ale > max_ale:
                max_ale = ale
    return max_ale


def mean_squared_error(list1,list2):
    """Compute the MSE, list 1 are true values, list 2 are predicted values."""
    return sklearn_mean_squared_error(list1,list2)


def rescale(value, source_from, source_to, target_from=0, target_to=1, how='linear'):
    """Rescale a value from one range to another."""

    if value < 0:
        raise ValueError('Cannot rescale negative value')
    if source_from < 0:
        raise ValueError('Cannot rescale using negative source_from')
    if source_to < 0:
        raise ValueError('Cannot rescale using negative source_to')
    if target_from < 0:
        raise ValueError('Cannot rescale using negative target_from')
    if target_to < 0:
        raise ValueError('Cannot rescale using negative target_to')
    if value < source_from:
        raise ValueError('Cannot rescale a value outside the source interval (value={}, source_from={}'.format(value, source_from))
    if value > source_to:
        raise ValueError('Cannot rescale a value outside the source interval (value={}, source_to={}'.format(value, source_to))

    source_total = source_to - source_from
    source_segment = value - source_from
    value_ratio = source_segment/source_total

    if target_from==0 and target_to==1:
        return value_ratio
    else:
        target_total = target_to - target_from
        return ((value_ratio*target_total)+target_from)

    # If how=log...log_compression /log base

def os_shell(command, capture=False, verbose=False, interactive=False, silent=False):
    """Execute a command in the OS shell. By default prints everything. If the capture switch is set,
    then it returns a namedtuple with stdout, stderr, and exit code."""

    if capture and verbose:
        raise Exception('You cannot ask at the same time for capture and verbose, sorry')

    # Log command
    logger.debug('Shell executing command: "%s"', command)

    # Execute command in interactive mode
    if verbose or interactive:
        exit_code = subprocess.call(command, shell=True)
        if exit_code == 0:
            return True
        else:
            return False

    # Execute command getting stdout and stderr
    # http://www.saltycrane.com/blog/2008/09/how-get-stdout-and-stderr-using-python-subprocess-module/

    process          = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    (stdout, stderr) = process.communicate()
    exit_code        = process.wait()

    # Convert to str (Python 3)
    stdout = stdout.decode(encoding='UTF-8')
    stderr = stderr.decode(encoding='UTF-8')

    # Formatting..
    stdout = stdout[:-1] if (stdout and stdout[-1] == '\n') else stdout
    stderr = stderr[:-1] if (stderr and stderr[-1] == '\n') else stderr

    # Output namedtuple
    Output = namedtuple('Output', 'stdout stderr exit_code')

    if exit_code != 0:
        if capture:
            return Output(stdout, stderr, exit_code)
        else:
            string  = '\n#---------------------------------'
            string += '\n# Shell exited with exit code {}'.format(exit_code)
            string += '\n#---------------------------------\n'
            string += '\nStandard output: "'
            string += stdout.encode("utf-8", errors="ignore")
            string += '"\n\nStandard error: "'
            string += stderr.encode("utf-8", errors="ignore") +'"\n\n'
            string += '#---------------------------------\n'
            string += '# End Shell output\n'
            string += '#---------------------------------\n'
            print(string)
            return False
    else:
        if capture:
            return Output(stdout, stderr, exit_code)
        elif not silent:
            # Just print stdout and stderr cleanly
            print(stdout)
            print(stderr)
            return True
        else:
            return True


class DistributionFunction():
    """A class representing a statistical distribution."""

    def __init__(self, dist, params):
        self.dist = dist
        self.params = params

        # Init distirbution object
        self.dist_obj = getattr(scipy.stats, self.dist)

        # Check correct params
        self(0)

    def __call__(self, x):
        return self.dist_obj.pdf(x, **self.params)

    def plot(self, x_min=-1, x_max=1, show=True):
        """Plot the distribution"""

        plt.clf()

        # Populate data
        X = np.linspace(x_min, x_max, 1000)
        Y = [self(x) for x in X]
        plt.plot(X, Y, 'k', linewidth=2, label=str(self.dist), color='orange')

        # Viz
        plt.legend(loc="upper right")
        plt.grid()
        #plt.title('Distribution function: {}'.format(self.dist))

        # Show or return
        if show:
            plt.show()
        else:
            return plt

    def find_x(self, y, wideness=1000, side='right'):
        # Note: "self" can be any function, we use this class as a callable here

        start = self.params['loc']
        if side == 'right':
            end = self.params['scale']*wideness
        elif side == 'left':
            end = -(self.params['scale']*wideness)
        else:
            raise ValueError('unknown side "{}"'.format(side))

        def scaled_gaussian(x):
            return self(x)-y

        x = optimize.bisect(scaled_gaussian, start, end)

        return x

#===========================
#  Private classes
#===========================

class _Gaussian():

    def __init__(self, mu, sigma, data=None):
        self.mu = mu
        self.sigma = sigma
        self.data = data

    def __call__(self, x):
        return stats.norm.pdf(x, self.mu, self.sigma)

    def cumulative(self,x):
        return stats.norm.cdf(x, self.mu, self.sigma)

    @classmethod
    def from_data(cls, data):
        mu, sigma = stats.norm.fit(data)
        return cls(mu,sigma,data)

    def plot(self, cumulative=False, bins=20):

        # Plot the histogram.
        if self.data:
            plt.hist(self.data, bins=bins, density=True, alpha=0.6, color='b')

        # Plot the gaussian itself
        xmin, xmax = plt.xlim()
        x = np.linspace(xmin, xmax, 100)

        if cumulative:
            title = 'Gasussian (mu={:.3f}, sigma={:.3f}) - cumulative'.format(self.mu, self.sigma)
            p = stats.norm.cdf(x, self.mu, self.sigma)
        else:
            title = 'Gasussian (mu={:.3f}, sigma={:.3f})'.format(self.mu, self.sigma)
            p = stats.norm.pdf(x, self.mu, self.sigma)

        plt.plot(x, p, 'k', linewidth=2)
        plt.title(title)

        # Show the plot
        plt.show()

        # Show the plot
        plt.show()


    def find_xes(self, y, wideness=1000):
        # Note: "self" can be any function, we use the gaussian class as a callable object here.

        start = self.mu
        end   = self.sigma*wideness

        def scaled_gaussian(x):
            return self(x)-y

        x = optimize.bisect(scaled_gaussian, start, end)

        return [ self.mu - (x-self.mu), x]


#===========================
#  Private utilities
#===========================

def _is_close(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def _is_almost_equal(one, two):
    if 0.95 < (one / two) <= 1.05:
        return True
    else:
        return False

def _set_from_t_and_to_t(from_dt, to_dt, from_t, to_t):
    """Set from_t and to_t from a (valid) combination of from_dt, to_dt, from_t, to_t."""

    # Sanity chceks
    if from_t is not None and not is_numerical(from_t):
        raise Exception('from_t is not numerical')
    if to_t is not None and not is_numerical(to_t):
        raise Exception('to_t is not numerical')
    if from_dt is not None and not isinstance(from_dt, datetime):
        raise Exception('from_dt is not datetime')
    if to_dt is not None and not isinstance(to_dt, datetime):
        raise Exception('to_t is not datetime')

    if from_dt is not None:
        if from_t is not None:
            raise Exception('Got both from_t and from_dt, pick one')
        from_t = s_from_dt(from_dt)
    if to_dt is not None:
        if to_t is not None:
            raise Exception('Got both to_t and to_dt, pick one')
        to_t = s_from_dt(to_dt)
    return from_t, to_t


def _item_is_in_range(item, from_t, to_t):
    """Check if an item (*TimePoint or *TimeSlot) is within the specified time range"""

    # TODO: maybe use a custom iterator for looping over time series items?
    # see https://stackoverflow.com/questions/6920206/sending-stopiteration-to-for-loop-from-outside-of-the-iterator
    from .datastructures import Point, Slot
    if isinstance(item, Slot):
        if from_t is not None and to_t is not None and from_t > to_t:
            if item.end.t <= to_t:
                return True
            if item.start.t >= from_t:
                return True
        else:
            if from_t is not None and item.start.t < from_t:
                return False
            if to_t is not None and item.end.t > to_t:
                raise StopIteration
            return True
    elif isinstance(item, Point):
        if from_t is not None and to_t is not None and from_t > to_t:
            if item.t <= to_t:
                return True
            if item.t >= from_t:
                return True
        else:
            if from_t is not None and item.t < from_t:
                return False
            if to_t is not None and item.t > to_t:
                raise StopIteration
            return True
    else:
        raise ConsistencyException('Got unknown type "{}'.format(item.__class__.__name__))


def _compute_validity_regions(series, from_t=None, to_t=None, sampling_interval=None, cut=False):
    """
    Compute the validity regions for the series points. If from_t or to_t are given, computes them within that interval only.

    Args:
        series: the time series.
        from_t: the interval start, if any.
        to_t: the interval end, if any.
        sampling_interval: the sampling interval. In not set, it will be used the series' (auto-detected) one.
        cut: if to shrink the validity according to the start/end.

    Returns:
        dict: for each point, the validity region start and end.
    """

    # If single-element series, check that the sampling_interval was given
    if len(series) == 1:
        if sampling_interval is None:
            raise ValueError('The series has only one element and no sampling_interval is provided, no idea how to compute validity')

    # Get the series sampling interval unless it is forced to a specific value
    if not sampling_interval:
        sampling_interval = series._autodetected_sampling_interval

    # Segments dict
    validity_segments = {}

    # Compute validity segments for each point
    prev_point_t = None
    for point in series:

        # Get point validity boundaries
        point_valid_from_t = point.t - (sampling_interval/2)
        point_valid_to_t   = point.t + (sampling_interval/2)

        # Are we processing points not belonging to this interval, if set?
        if from_t is not None and point_valid_to_t <= from_t:
            continue
        if to_t is not None and point_valid_from_t > to_t:
            break

        # Shrink with respect to start-end if required
        if cut:
            if from_t is not None and point_valid_from_t < from_t:
                point_valid_from_t = from_t
            if to_t is not None and point_valid_to_t > to_t:
                point_valid_to_t = to_t

        # Shrink if overlaps: if the previous point validity overlap with ours, resize both
        if prev_point_t is not None:
            if point_valid_from_t < validity_segments[prev_point_t][1]:
                # Shrink both
                mid_t = (validity_segments[prev_point_t][1] + point_valid_from_t)/2
                validity_segments[prev_point_t][1] = mid_t
                point_valid_from_t = mid_t

        # Set this validity segment boundaries
        validity_segments[point.t] = [point_valid_from_t,point_valid_to_t]

        # Set previous point timestamp
        prev_point_t = point.t

    return validity_segments


def _compute_coverage(series, from_t, to_t, sampling_interval=None):
    """
    Compute the coverage of an interval based on the points validity regions.

    Args:
        series: the time series.
        from_t: the interval start.
        to_t: the interval end.
        sampling_interval: the sampling interval. In not set, it will be used the series' (auto-detected) one.

    Returns:
        float: the interval coverage.
    """

    logger.debug('Called _compute_coverage() from {} to {}'.format(from_t, to_t))

    # Check from_t/to_t
    if from_t >= to_t:
        raise ValueError('From is higher than to (or equal): from_t="{}", to_t="{}"'.format(from_t, to_t))

    # If the series is empty, return zero coverage
    if not series:
        return 0.0

    # Compute the validity regions in the interval. if point haven't them?
    # validity_regions = _compute_validity_regions(series, from_t=from_t, to_t=to_t, cut=True, sampling_interval=sampling_interval)
    # And now attach?

    # Sum all the validity regions according to the from/to:
    coverage_s = 0
    for point in series:
        try:
            # Skip at the beginning and end if any
            if point.valid_to < from_t:
                continue
            if point.valid_from >= to_t:
                break

            # Skip if interpolated as well
            try:
                if point._interpolated:
                    continue
            except AttributeError:
                pass

            # Set this point valid from/to accoring to the interval boundaries
            this_point_valid_from = point.valid_from if point.valid_from >= from_t else from_t
            this_point_valid_to = point.valid_to if point.valid_to < to_t else to_t

            # Add this point coverage to the overall coverage
            coverage_s += this_point_valid_to-this_point_valid_from

            logger.debug('Point %s: from %s, to %s (current total=%s) ', point.t, this_point_valid_from, this_point_valid_to, coverage_s)

        except AttributeError:
            raise AttributeError('Point {} has no valid_to or valid_from'.format(point)) from None

    # Convert from coverage in seconds to ratio:
    coverage = coverage_s / (to_t - from_t)

    # Return
    logger.debug('_compute_coverage: Returning %s (%s percent)', coverage, coverage*100.0)
    return coverage


def _compute_data_loss(series, from_t, to_t, force=False, sampling_interval=None):
    """
    Compute the data loss  of an interval based on the points validity regions.

    Args:
        series: the time series.
        from_t: the interval start.
        to_t: the interval end.
        force: if to force computing even when not strictly necessary.
        sampling_interval: the sampling interval. In not set, it will be used the series' (auto-detected) one.

    Returns:
        float: the interval data loss.
    """

    logger.debug('Called _compute_data_loss() from {} to {}'.format(from_t, to_t))

    # Get the series sampling interval unless it is forced to a specific value
    if not sampling_interval:
        sampling_interval = series._autodetected_sampling_interval

    # Compute the data loss from missing coverage.
    # Note: to improve performance, this could be computed only if the series is "raw" or forced,
    # i.e. at first and last items since on the borders there still may be  data losses.
    data_loss_from_missing_coverage = 1 - _compute_coverage(series, from_t, to_t, sampling_interval=sampling_interval)

    logger.debug('Data loss from missing coverage: %s', data_loss_from_missing_coverage)

    # Compute the data loss from previous data losses. Computed only if the resolutions is constant, as they can be defined only
    # if the series was already  made uniform (i.e. by a resampling process), or if forced for testing or other potential reasons.
    data_loss_from_previously_computed = 0.0

    if (series.resolution is not None) or force:

        for point in series:

            if point.data_loss:

                # Set validity
                point_valid_from_t = point.t - (sampling_interval/2)
                point_valid_to_t = point.t + (sampling_interval/2)

                # Skip points which have not to be taken into account
                if point_valid_to_t < from_t:
                    continue
                if point_valid_from_t >= to_t:
                    continue

                # Compute the contribution of this point data loss: overlapping
                # first point, overlapping last point, or all included?
                if point_valid_from_t < from_t:
                    point_contribution = abs(from_t - point_valid_to_t)
                elif point_valid_to_t >= to_t:
                    point_contribution = abs(to_t - point_valid_from_t)
                else:
                    point_contribution = sampling_interval

                #logger.debug('----------------')
                #count+=1
                #logger.debug('count: %s', count)
                #logger.debug('point_valid_from_t: %s', point_valid_from_t)
                #logger.debug('point.t: %s', point.t)
                #logger.debug('point_valid_to_t: %s', point_valid_to_t)
                #logger.debug('point_contribution: %s', point_contribution)

                # Now rescale the data loss with respect to the contribution and from/to and sum it up
                data_loss_from_previously_computed += point.data_loss * (point_contribution/( to_t - from_t))

    logger.debug('Data loss from previously computed: %s', data_loss_from_previously_computed)

    # Compute total data loss
    data_loss = data_loss_from_missing_coverage + data_loss_from_previously_computed

    # The next step is controversial, as it will cause to abuse the "None" data losses that
    # are only used for the forecasts at the moment. TODO: what do we want to do here?
    #if series_resolution != 'variable' and not data_loss:
    #    data_loss = None

    # Return
    return data_loss


def _check_timeseries(series):
    """Check a time series for type, not emptiness and fixed resolution."""

    # Import here or you will end up with cyclic imports
    from .datastructures import TimeSeries

    if not isinstance(series, TimeSeries):
        raise TypeError('A TimeSeries object is required (got "{}")'.format(series.__class__.__name__))

    if not series:
        raise ValueError('A non-empty time series is required')

    if series.resolution == 'variable':
        raise ValueError('Time series with undefined (variable) resolutions are not supported. Resample or slot the time series first.')


def _check_resolution(series, resolution):

    def __check_resolution(series, resolution):
        # TODO: Fix this mess.. Make the .resolution behavior consistent!
        if resolution == series.resolution:
            return True
        try:
            if resolution == series.resolution.as_seconds():
                return True
        except:
            pass
        try:
            if resolution.as_seconds() == series.resolution:
                return True
        except:
            pass
        try:
            if resolution.as_seconds() == series.resolution:
                return True
        except:
            pass
        return False

    # Check series resolution
    if not __check_resolution(series, resolution):
        raise ValueError('This model is fitted on "{}" resolution data, while your data has "{}" resolution.'.format(resolution, series.resolution))

def _check_data_labels(series, data_labels):
    series_data_labels = series.data_labels()
    if len(series_data_labels) != len(data_labels):
        raise ValueError('This model is fitted on {} data labels, while your data has {} data labels.'.format(len(data_labels), len(series_data_labels)))
    if series_data_labels != data_labels:
        # TODO: logger.warning?
        raise ValueError('This model is fitted on "{}" data labels, while your data has "{}" data labels.'.format(data_labels, series_data_labels))

def _check_series_of_points_or_slots(series):
    from .datastructures import DataPoint, DataSlot
    if not (issubclass(series.item_type, DataPoint) or issubclass(series.item_type, DataSlot)):
        raise TypeError('Cannot operate on a series of "{}", only series of DataPoints or DataSlots are supported'.format(series.item_type.__name__))

def _check_indexed_data(series):
    try:
        series.data_labels()
    except TypeError:
        raise TypeError('Cannot operate on a series of "{}" with "{}" data, only series of indexed data (as lists and dicts) are supported'.format(series.item_type.__name__, series[0].data.__class__.__name__))

def _get_periodicity_index(item, resolution, periodicity, dst_affected=False):
    """Get the periodicty index."""

    from .units import Unit, TimeUnit
    # Handle specific cases
    if isinstance(resolution, TimeUnit):
        resolution_s = resolution.as_seconds(item.dt)
    elif isinstance(resolution, Unit):
        if isinstance(resolution, list):
            raise NotImplementedError('Sorry, periodocity in multi-dimensional spaces are not defined')
        resolution_s = resolution
    else:
        if isinstance(resolution, list):
            raise NotImplementedError('Sorry, periodocity in multi-dimensional spaces are not defined')
        resolution_s = resolution

    # Compute periodicity index
    if not dst_affected:

        # Get index based on item timestamp, normalized to unit, modulus periodicity
        periodicity_index =  int(item.t / resolution_s) % periodicity

    else:

        # Get periodicity based on the datetime

        # Do we have an active DST?
        dst_timedelta = item.dt.dst()

        if dst_timedelta.days == 0 and dst_timedelta.seconds == 0:
            # No DST
            periodicity_index = int(item.t / resolution_s) % periodicity

        else:
            # DST
            if dst_timedelta.days != 0:
                raise Exception('Don\'t know how to handle DST with days timedelta = "{}"'.format(dst_timedelta.days))

            if resolution_s > 3600:
                raise Exception('Sorry, this time series has not enough resolution to account for DST effects (resolution_s="{}", must be below 3600 seconds)'.format(resolution_s))

            # Get DST offset in seconds
            dst_offset_s = dst_timedelta.seconds # 3600 usually

            # Compute the periodicity index
            periodicity_index = (int((item.t + dst_offset_s) / resolution_s) % periodicity)

    return periodicity_index


def _sanitize_string(string, no_data_placeholders=[]):
    """Sanitize the encofing of a stringwhile handling no data placeholders"""

    string = re.sub('\s+',' ',string).strip()
    if string.startswith('\'') or string.startswith('"'):
        string = string[1:]
    if string.endswith('\'') or string.endswith('"'):
        string = string[:-1]
    string = string.strip()
    if string.lower().replace('.','') in no_data_placeholders:
        return None
    return string


def _is_list_of_integers(the_list):
    """Check if the list if made of integers"""

    for item in the_list:
        if not isinstance(item, int):
            return False
    else:
        return True


def _to_float(string,no_data_placeholders=[],label=None):
    """Convert to float while handling no data placeholders and discarding data indexes"""

    sanitized_string_string = _sanitize_string(string,no_data_placeholders)
    if sanitized_string_string:
        sanitized_string_string = sanitized_string_string.replace(',','.')
    try:
        return float(sanitized_string_string)
    except (ValueError, TypeError):
        # Do not raise inf converting indexes as they are allowed to be "None"
        if label and label.startswith('__'):
            return None
        raise FloatConversionError(sanitized_string_string)



def _to_time_unit_string(seconds, friendlier=True):
    """Converts seconds to a (friendlier) time unit string, as 1s, 1h, 10m etc.)"""

    seconds_str = str(seconds).replace('.0', '')
    if friendlier:
        if seconds_str == '60':
            seconds_str = '1m'
        elif seconds_str == '600':
            seconds_str = '10m'
        elif seconds_str == '3600':
            seconds_str = '1h'
        else:
            seconds_str = seconds_str+'s'
    else:
        seconds_str = seconds_str+'s'
    return seconds_str


def _compute_distribution_approximation_errors(distribution_function, prediction_errors, bins=30, details=False):

    # Support vars
    max_prediction_error = max(prediction_errors)
    min_prediction_error = min(prediction_errors)
    error_unit = max_prediction_error - min_prediction_error

    # Create the bins
    error_step = error_unit / bins
    prediction_error_bins = [[] for _ in range(bins)]
    for i, prediction_error in enumerate(prediction_errors):
        for j in range(bins-1):
            if ( (min_prediction_error + (error_step*j))  <= prediction_error <= (min_prediction_error + (error_step*(j+1))) ):
                prediction_error_bins[j].append(prediction_error)

    real_distribution_values = {} # x -> y
    max_len = 0
    for i, prediction_error_bin in enumerate(prediction_error_bins):
        if prediction_error_bin:
            real_distribution_values[min_prediction_error + (error_step*i) + (error_step/2)] = len(prediction_error_bin)
            if len(prediction_error_bin) > max_len:
                max_len = len(prediction_error_bin)

    # Normalize
    for key in real_distribution_values:
        real_distribution_values[key] = real_distribution_values[key]/max_len

    # Compute the approximation errors
    approximation_errors = []
    binned_distribution_values = []
    binned_real_distribution_values = []
    for key in real_distribution_values:
        #logger.debug('{}: \t{}\t vs\t {}'.format(key, distribution(key),real_distribution_values[key]))
        binned_distribution_values.append(distribution_function(key))
        binned_real_distribution_values.append(real_distribution_values[key])
        approximation_errors.append(abs(distribution_function(key)-real_distribution_values[key]))

    if details:
        return (approximation_errors, binned_distribution_values, binned_real_distribution_values)
    else:
        return approximation_errors


def _detect_notebook_major_version():
    stdout = subprocess.check_output(["jupyter", "--version"]).decode()
    versions={}
    for line in stdout.split('\n'):
        if ':' in line:
            line=line.strip()
            what, version = line.split(':')
            versions[what.strip()] = version.strip()
    #print(versions)
    if not versions or 'notebook' not in versions:
        notebook_major_version = None
    else:
        notebook_major_version = int(versions['notebook'].split('.')[0])
    return notebook_major_version


def _is_index_based(data):
    try:
        for i in range(len(data)):
            data[i]
    except:
        return False
    else:
        return True


def _is_key_based(data):
    if _is_index_based(data):
        return False
    try:
        for key in data:
            data[key]
    except:
        return False
    else:
        return True


def _has_numerical_values(data):

    if _is_index_based(data):
        for i in range(len(data)):
            try:
                data[i] + 1
            except:
                return False
        return True

    elif _is_key_based(data):
        for key in data:
            try:
                data[key] + 1
            except:
                return False
        return True

    else:
        raise ValueError('Don\'t know how to establish if this data type has numerical values or not (got "{}")'.format(data.__class__.__name__))




import unittest
import os
import tempfile
from ..datastructures import DataTimePoint, DataTimeSlot, TimeSeries
from ..storages import CSVFileStorage
from ..units import TimeUnit

# Setup logging
from .. import logger
logger.setup()

# Set test data path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'


class TestCSVFileStorage(unittest.TestCase):

    def test_CSVFileStorage_get_basic(self):

        # Basic iso8601 with two columns, one for the timestamp and one for the value, no labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/single_value_no_labels.csv')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 6)
        self.assertEqual(timeseries[0].t, 946684800)
        self.assertEqual(timeseries[0].data, [1000])

        # Basic iso8601 multi values no labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_no_labels.csv')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, [1000,10])
        self.assertEqual(timeseries[-1].data, [1040,14])

        # Basic iso8601 multi values no labels and filtering
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels=[2])
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, [10])
        self.assertEqual(timeseries[-1].data, [14])

        # Basic iso8601 multi values no labels and filtering and force data type
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels=[1], data_type=float)
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, [1000])
        self.assertEqual(timeseries[-1].data, [1040])

        # Basic iso8601 multi values with labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, {'flow': 1000.0, 'temp': 10.0})
        self.assertEqual(timeseries[-1].data, {'flow': 1040.0, 'temp': 14.0})

        # Test we don't have any extra data index as well
        self.assertEqual(timeseries._all_data_indexes(), [])

        # Basic iso8601 multi values with labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels=['temp'])
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, {'temp': 10.0})
        self.assertEqual(timeseries[-1].data,{'temp': 14.0})

        # Basic iso8601 multi values with labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels=[1])
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 50)
        self.assertEqual(timeseries[0].data, [1000.0])
        self.assertEqual(timeseries[-1].data, [1040.0])


    def test_CSVFileStorage_get_errors(self):

        # Test wrong datatypes:
        with self.assertRaises(Exception):
            CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels = 0)
        with self.assertRaises(Exception):
            CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels = 'flow')
        with self.assertRaises(Exception):
            CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', time_label = [0])
        with self.assertRaises(Exception):
            CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', date_label = [0])

        # Test requiring a not existent data column:
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', data_labels = ['temp', 'flow_NO'])
        with self.assertRaises(Exception):
            storage.get()


    def test_CSVFileStorage_get_dirty(self):

        # Basic iso8601 with two columns, one for the timestamp and one for the value, no labels, dirty
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/single_value_no_labels_dirty.csv')
        with self.assertRaises(Exception):
            _ = storage.get()

        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/single_value_no_labels_dirty.csv', skip_errors=True)
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 3)
        self.assertEqual(timeseries[0].t, 946684800)
        self.assertEqual(timeseries[0].data, [1000])

        # Basic iso8601 with two labels, dirty
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels_dirty.csv', separator=';', skip_errors=True)
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 3)
        self.assertEqual(timeseries[0], DataTimePoint(t= 946688400, data={'flow': 1010.0, 'temp': 11.0}))
        self.assertEqual(timeseries[1], DataTimePoint(t= 946706400, data={'flow': 1100.0, 'temp': 10.7}))
        self.assertEqual(timeseries[2], DataTimePoint(t= 946713600, data={'flow': 1300.0, 'temp': 10.0}))


    def test_CSVFileStorage_get_unordered(self):

        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels_unordered.csv', sort=True)
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 5)
        self.assertEqual(timeseries[0].t, 946684800)
        self.assertEqual(timeseries[0].data, {'flow': 1000.0, 'temp': 10.0})
        self.assertEqual(timeseries[-1].t, 946684800 + 60*60*4)
        self.assertEqual(timeseries[-1].data, {'flow': 1040.0, 'temp': 14.0})


    def test_CSVFileStorage_get_timestamp_formats(self):

        # Autodetect epoch
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/temp_long_10m.csv')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 19140)

        # Autodetect iso
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv')
        self.assertEqual(storage.timestamp_format, 'auto')
        self.assertEqual(storage.timestamp_column, 'auto')
        timeseries = storage.get()
        self.assertEqual(storage.timestamp_format, 'iso8601')
        self.assertEqual(storage.timestamp_column, 'timestamp')
        self.assertEqual(len(timeseries), 50)

        # Epoch timestamp format
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/temp_short_1h.csv',
                                  timestamp_column = 'epoch',
                                  timestamp_format = 'epoch',)
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 100)
        self.assertEqual(timeseries[0].t, 1546477200)
        self.assertEqual(timeseries[-1].t, 1546833600)

        # Use only month and year as date column
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/shampoo_sales.csv',
                                 date_column = 'Month',
                                 date_format = '%y-%m')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 36)
        self.assertEqual(timeseries[0].t, 978307200)
        self.assertEqual(timeseries[-1].t, 1070236800)

        # Separate time and date columns, custom format
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format2.csv',
                                 date_column = 'Date',
                                 date_format = '%d/%m/%Y',
                                 time_column = 'Time',
                                 time_format = '%H:%M')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 6)
        self.assertEqual(timeseries[0].t, 1583280000)
        self.assertEqual(timeseries[-1].t, 1583298000)

        # Test only date column and without meaningful timestamp label (and force points)
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/only_date_no_meaningful_timestamp_label.csv',
                                 timestamp_format = '%Y-%m-%d', series_type='points')

        timeseries = storage.get()
        self.assertEqual(len(timeseries), 95)
        self.assertEqual(timeseries[0].t, 1197244800)
        self.assertEqual(timeseries[-1].t, 1205798400)

        # Test data with quotes
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format5.csv')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 6)
        self.assertEqual(timeseries[2].data['temp'], 23.34)
        self.assertEqual(timeseries[3].data['humi'], 55)

        # Test getting only a specific data labels
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format5.csv')
        timeseries = storage.get(filter_data_labels=['temp'])
        self.assertEqual(len(timeseries), 6)
        self.assertEqual(timeseries[0].data_labels(), ['temp'])


    def test_CSVFileStorage_get_timezones(self):

        # Set the default timezone
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/multi_values_with_labels.csv', tz='Europe/Rome')
        timeseries = storage.get()
        self.assertEqual(timeseries[0].t, 946684800.0)
        self.assertEqual(str(timeseries[0].tz), 'Europe/Rome')

        # Force a a specific timezone
        timeseries = storage.get(force_tz='America/New_York')
        self.assertEqual(timeseries[0].t, 946684800.0)
        self.assertEqual(str(timeseries.tz), 'America/New_York')

        # Handle naive timestamps (04/03/2020 00:00)
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format3.csv',
                                 timestamp_format = '%d/%m/%Y %H:%M',
                                 tz='Europe/Rome')
        timeseries = storage.get()
        self.assertEqual(timeseries[0].t, 1583276400.0)
        self.assertEqual(str(timeseries.tz), 'Europe/Rome')

        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format3.csv',
                                 timestamp_format = '%d/%m/%Y %H:%M')
        timeseries = storage.get(force_tz='Europe/Rome')
        self.assertEqual(timeseries[0].t, 1583276400.0)
        self.assertEqual(str(timeseries.tz), 'Europe/Rome')

        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/format3.csv',
                                 timestamp_format = '%d/%m/%Y %H:%M',
                                 tz='Europe/Rome')
        timeseries = storage.get(force_tz='America/New_York')
        self.assertEqual(timeseries[0].t, 1583298000.0)
        self.assertEqual(str(timeseries.tz), 'America/New_York')


    def test_CSVFileStorage_get_slots_reconstruction(self):

        # Let the storage to generate slots with interpolated data
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/only_date_no_meaningful_timestamp_label.csv',
                                 timestamp_format = '%Y-%m-%d')
        timeseries = storage.get()
        self.assertEqual(len(timeseries), 100)

        # Check first
        self.assertEqual(timeseries[0].t, 1197244800)

        # Test correct data loss set and reconstruction
        self.assertEqual(timeseries[52].data_loss, 1)
        self.assertAlmostEqual(timeseries[52].data['y'], 8.748874781508915)

        self.assertEqual(timeseries[80].data_loss, 1)
        self.assertAlmostEqual(timeseries[80].data['y'], 7.56004994651234)

        self.assertEqual(timeseries[82].data_loss, 1)
        self.assertAlmostEqual(timeseries[82].data['y'], 7.444587100634211)

        self.assertEqual(timeseries[84].data_loss, 1)
        self.assertAlmostEqual(timeseries[84].data['y'], 7.599538949266937)

        self.assertEqual(timeseries[85].data_loss, 1)
        self.assertAlmostEqual(timeseries[85].data['y'], 7.862140984826253)

        # Check last
        self.assertEqual(timeseries[-1].t, 1205798400)

        # Check all others have data_loss = None
        for i, item in enumerate(timeseries):
            if i not in [52, 80, 82, 84, 85]:
                self.assertEqual(item.data_loss, None, 'Failed for i={}'.format(i))
            else:
                self.assertEqual(item.data_loss, 1, 'Failed for i={}'.format(i))

    def test_CSVFileStorage_get_series_types(self):

        # Get data as slots
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/temp_short_1h.csv')
        timeseries = storage.get(force_slots=True)
        self.assertEqual(len(timeseries), 100)
        self.assertEqual(timeseries[0].start.t, 1546477200)
        self.assertEqual(timeseries[-1].start.t, 1546833600)


    def test_CSVFileStorage_get_no_data(self):
        storage = CSVFileStorage(TEST_DATA_PATH + '/csv/no_data.csv')
        from ..exceptions import NoDataException
        with self.assertRaises(NoDataException):
            storage.get()


    def test_CSVFileStorage_put(self):

        with tempfile.TemporaryDirectory() as temp_dir:

            # Test on points
            timeseries = TimeSeries(DataTimePoint(t=60, data=[23.8,3], data_loss=0.1),
                                    DataTimePoint(t=120, data=[24.1,4], data_loss=0.2),
                                    DataTimePoint(t=240, data=[23.1,5], data_loss=0.3),
                                    DataTimePoint(t=300, data=[22.7,6], data_loss=0.4))
            timeseries.change_tz('Europe/Rome')

            storage = CSVFileStorage('/{}/file_1.csv'.format(temp_dir))
            storage.put(timeseries)

            with self.assertRaises(Exception):
                storage.put(timeseries)

            storage.put(timeseries, overwrite=True)

            # Test getting back the series from the storage
            series = storage.get()
            self.assertEqual(len(series),4)
            self.assertTrue(isinstance(series[0], DataTimePoint))

            # Test indexes correctly se by the get
            self.assertEqual(len(series[0].data_indexes),1)
            self.assertEqual(series[0].data_indexes['data_loss'], 0.1)
            self.assertEqual(series[3].data_indexes['data_loss'], 0.4)

            # Now test on slots
            timeseries = TimeSeries(DataTimeSlot(t=60, unit=TimeUnit('1m'), data=[23.8,3], data_loss=0.1),
                                    DataTimeSlot(t=120, unit=TimeUnit('1m'), data=[24.1,4], data_loss=0.2),
                                    DataTimeSlot(t=180, unit=TimeUnit('1m'), data=[23.1,5], data_loss=0.3),
                                    DataTimeSlot(t=240, unit=TimeUnit('1m'), data=[22.7,6], data_loss=0.4))
            timeseries.change_tz('Europe/Rome')

            storage = CSVFileStorage('/{}/file_2.csv'.format(temp_dir))
            storage.put(timeseries)

            with self.assertRaises(Exception):
                storage.put(timeseries)

            storage.put(timeseries, overwrite=True)

            # Test getting back the series from the storage
            series = storage.get(force_slots=True)
            self.assertEqual(len(series),4)
            self.assertTrue(isinstance(series[0], DataTimeSlot))

            # Test indexes correctly se by the get
            self.assertEqual(len(series[0].data_indexes),1)
            self.assertEqual(series[0].data_indexes['data_loss'], 0.1)
            self.assertEqual(series[3].data_indexes['data_loss'], 0.4)





import unittest
import os
import tempfile
from propertime.utils import dt

from ..datastructures import TimePoint, DataTimeSlot, DataTimePoint, TimeSeries
from ..models.base import Model, _KerasModel
from ..exceptions import NotFittedError

# Setup logging
from .. import logger
logger.setup()

# Test data and temp path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'
TEMP_MODELS_DIR = tempfile.TemporaryDirectory().name

# Ensure reproducibility
import random
import numpy as np
random.seed(0)
np.random.seed(0)



class TestBaseBaseModel(unittest.TestCase):

    def test_Model(self):

        # Test non-parametric
        model = Model()
        self.assertFalse(model._is_parametric())
        self.assertEqual(model._type, 'non-parametric')
        with self.assertRaises(TypeError):
            model.save(TEMP_MODELS_DIR+'/test_model')
        with self.assertRaises(TypeError):
            model.id

        # Test parametric
        class ParametricModelMock(Model):
            def __init__(self, param1=0):
                self.data = {}
                self.data['param1'] = param1
                super(ParametricModelMock, self).__init__()

        parametric_model = ParametricModelMock(param1=1)
        self.assertTrue(parametric_model._is_parametric())
        self.assertEqual(parametric_model._type, 'parametric')
        self.assertEqual(parametric_model.data['param1'], 1)
        parametric_model.save(TEMP_MODELS_DIR+'/parametric_model')

        parametric_model = ParametricModelMock.load(TEMP_MODELS_DIR+'/parametric_model')
        self.assertTrue(parametric_model._is_parametric())
        self.assertEqual(parametric_model.data['param1'], 1)



    def test_Model_with_TimeSeries(self):

        # Define test time series
        empty_timeseries = TimeSeries()
        timeseries = TimeSeries(DataTimeSlot(start=TimePoint(t=1), end=TimePoint(t=2), data={'metric1': 56}),
                                 DataTimeSlot(start=TimePoint(t=2), end=TimePoint(t=3), data={'metric1': 56}))

        # Define a fittable parametric model mock
        class FittableParametricModelMock(Model):
            @Model.fit_method
            def fit(self, series):
                self.data['param1'] = 1
            @Model.predict_method
            def predict(self, series):
                return series
            @Model.apply_method
            def apply(self, series):
                return series
            @Model.evaluate_method
            def evaluate(self, series):
                return {'grade': 'A'}

        parametric_model = FittableParametricModelMock()
        self.assertTrue(parametric_model._is_parametric())
        self.assertEqual(parametric_model._type, 'parametric')

        # Cannot predict form a model before fitting
        with self.assertRaises(NotFittedError):
            parametric_model.predict(timeseries)

        # Cannot apply model before fitting
        with self.assertRaises(NotFittedError):
            parametric_model.apply(timeseries)

        # Cannot evaluate model before fitting
        with self.assertRaises(NotFittedError):
            parametric_model.evaluate(timeseries)

        # Cannot save model before fitting
        with self.assertRaises(NotFittedError):
            parametric_model.save(TEMP_MODELS_DIR+'/parametric_model')

        # Fit the model
        with self.assertRaises(ValueError):
            parametric_model.fit(empty_timeseries)
        parametric_model.fit(timeseries)
        self.assertEqual(parametric_model.data['param1'], 1)

        # Check predict
        with self.assertRaises(ValueError):
            parametric_model.predict(empty_timeseries)
        with self.assertRaises(TypeError):
            parametric_model.predict([1,2,3,4,5,6])
        self.assertEqual(parametric_model.predict(timeseries), timeseries)

        # Check apply
        with self.assertRaises(ValueError):
            parametric_model.apply(empty_timeseries)
        with self.assertRaises(TypeError):
            parametric_model.apply([1,2,3,4,5,6])
        self.assertEqual(parametric_model.apply(timeseries), timeseries)

        # Check evaluate
        with self.assertRaises(ValueError):
            parametric_model.evaluate(empty_timeseries)
        with self.assertRaises(TypeError):
            parametric_model.evaluate([1,2,3,4,5,6])
        self.assertEqual(parametric_model.evaluate(timeseries), {'grade': 'A'})

        # Save and re-load
        parametric_model.save(TEMP_MODELS_DIR+'/fittable_parametric_model')
        loaded_parametric_model = FittableParametricModelMock.load(TEMP_MODELS_DIR+'/fittable_parametric_model')
        self.assertTrue(loaded_parametric_model._is_parametric())
        self.assertEqual(loaded_parametric_model.data['param1'], 1)
        self.assertEqual(loaded_parametric_model.id, parametric_model.id)


class TestBaseKerasModel(unittest.TestCase):

    def setUp(self):

        self.test_timeseries = TimeSeries(DataTimePoint(t=1, data={'label_1': 0.1}),
                                          DataTimePoint(t=2, data={'label_1': 0.2}),
                                          DataTimePoint(t=3, data={'label_1': 0.3}),
                                          DataTimePoint(t=4, data={'label_1': 0.4}),
                                          DataTimePoint(t=5, data={'label_1': 0.5}),
                                          DataTimePoint(t=6, data={'label_1': 0.6}),)

        self.test_timeseries_mv = TimeSeries(DataTimePoint(t=1, data={'label_1': 0.1, 'label_2': 1.0}),
                                             DataTimePoint(t=2, data={'label_1': 0.2, 'label_2': 2.0}),
                                             DataTimePoint(t=3, data={'label_1': 0.3, 'label_2': 3.0}),
                                             DataTimePoint(t=4, data={'label_1': 0.4, 'label_2': 4.0}),
                                             DataTimePoint(t=5, data={'label_1': 0.5, 'label_2': 5.0}),
                                             DataTimePoint(t=6, data={'label_1': 0.6, 'label_2': 6.0}),)

    def test_to_matrix_representation(self):

        window_elements_matrix, _, _ = _KerasModel._to_matrix_representation(series = self.test_timeseries,
                                                                             window = 2,
                                                                             steps = 1,
                                                                             context_data_labels = None,
                                                                             target_data_labels = self.test_timeseries.data_labels(),
                                                                             data_loss_limit = None)

        # What to expect (using the timestamp to represent a data point):
        # 1,2
        # 2,3
        # 3,4
        # 4,5

        self.assertEqual(len(window_elements_matrix), 4)

        self.assertEqual(window_elements_matrix[0][0].t, 1)
        self.assertEqual(window_elements_matrix[0][1].t, 2)

        self.assertEqual(window_elements_matrix[1][0].t, 2)
        self.assertEqual(window_elements_matrix[1][1].t, 3)

        self.assertEqual(window_elements_matrix[2][0].t, 3)
        self.assertEqual(window_elements_matrix[2][1].t, 4)

        self.assertEqual(window_elements_matrix[-1][0].t, 4)
        self.assertEqual(window_elements_matrix[-1][1].t, 5)

        _, target_values_vector, _ = _KerasModel._to_matrix_representation(series = self.test_timeseries,
                                                                           window = 2,
                                                                           steps = 1,
                                                                           context_data_labels = None,
                                                                           target_data_labels = ['label_1'],
                                                                           data_loss_limit = None)

        # What to expect (using the data value to represent a data point):
        # [0.3], [0.4], [0.5], [0.6] Note that they are lists in order to support multi-step forecast
        self.assertEqual(target_values_vector, [[0.3], [0.4], [0.5], [0.6]])

        # Test for multivariate as well
        _, target_values_vector, _ = _KerasModel._to_matrix_representation(series = self.test_timeseries_mv,
                                                                           window = 2,
                                                                           steps = 1,
                                                                           context_data_labels = self.test_timeseries_mv.data_labels(),
                                                                           target_data_labels = ['label_1'],
                                                                           data_loss_limit = None)


        self.assertEqual(target_values_vector, [[0.3], [0.4], [0.5], [0.6]])


    def test_compute_window_features(self):

        window_elements_matrix, _, _ = _KerasModel._to_matrix_representation(series = self.test_timeseries,
                                                                             window = 3,
                                                                             steps = 1,
                                                                             context_data_labels = None,
                                                                             target_data_labels = self.test_timeseries.data_labels(),
                                                                             data_loss_limit = None)



        window_features  = _KerasModel._compute_window_features(window_elements_matrix[0],
                                                                data_labels = self.test_timeseries.data_labels(),
                                                                time_unit = self.test_timeseries.resolution,
                                                                features = ['values'])

        self.assertEqual(window_features, [[0.1],[0.2],[0.3]])


        window_features  = _KerasModel._compute_window_features(window_elements_matrix[0],
                                                                data_labels = self.test_timeseries.data_labels(),
                                                                time_unit = self.test_timeseries.resolution,
                                                                features = ['values', 'diffs'])
        self.assertAlmostEqual(window_features[0][0], 0.1)
        self.assertAlmostEqual(window_features[0][1], 0.1)
        self.assertAlmostEqual(window_features[1][0], 0.2)
        self.assertAlmostEqual(window_features[1][1], 0.1)
        self.assertAlmostEqual(window_features[-1][0], 0.3)
        self.assertAlmostEqual(window_features[-1][1], 0.1)

        # Multivariate

        window_elements_matrix, _, _ = _KerasModel._to_matrix_representation(series = self.test_timeseries_mv,
                                                                             window = 3,
                                                                             steps = 1,
                                                                             context_data_labels = None,
                                                                             target_data_labels = self.test_timeseries_mv.data_labels(),
                                                                             data_loss_limit = None)

        window_features  = _KerasModel._compute_window_features(window_elements_matrix[0],
                                                                data_labels = self.test_timeseries_mv.data_labels(),
                                                                time_unit = self.test_timeseries_mv.resolution,
                                                                features = ['values'])

        self.assertEqual(window_features, [[0.1, 1.0], [0.2, 2.0], [0.3, 3.0]])

        # Multivariate with context
        window_features  = _KerasModel._compute_window_features(window_elements_matrix[0],
                                                                data_labels = self.test_timeseries_mv.data_labels(),
                                                                time_unit = self.test_timeseries_mv.resolution,
                                                                features = ['values', 'diffs', 'hours'],
                                                                context_data = {'label_2':4.0})
        self.assertAlmostEqual(window_features[0][0], 0.1)
        self.assertAlmostEqual(window_features[0][1], 1.0)
        self.assertAlmostEqual(window_features[0][2], 0.1)
        self.assertAlmostEqual(window_features[0][3], 1.0)
        self.assertAlmostEqual(window_features[0][4], 0.0)

        self.assertAlmostEqual(window_features[1][0], 0.2)
        self.assertAlmostEqual(window_features[1][1], 2.0)
        self.assertAlmostEqual(window_features[1][2], 0.1)
        self.assertAlmostEqual(window_features[1][3], 1.0)
        self.assertAlmostEqual(window_features[1][4], 0.0)

        self.assertAlmostEqual(window_features[-1][0], 0.0)
        self.assertAlmostEqual(window_features[-1][1], 4.0)
        self.assertAlmostEqual(window_features[-1][2], 0.0)
        self.assertAlmostEqual(window_features[-1][3], 1.0)
        self.assertAlmostEqual(window_features[-1][4], 0.0)




import unittest
from propertime.utils import dt
from ..datastructures import TimeSeries, DataTimePoint

# Setup logging
from .. import logger
logger.setup()


class TestPlots(unittest.TestCase):

    def test_basic(self):

        timeseries = TimeSeries(DataTimePoint(dt=dt(2015,10,24,0,0,0, tz='Europe/Rome'), data={'value':23.8}),
                                DataTimePoint(dt=dt(2015,10,25,0,0,0, tz='Europe/Rome'), data={'value':22.1}),
                                DataTimePoint(dt=dt(2015,10,26,0,0,0, tz='Europe/Rome'), data={'value':25.3}))

        html_output = timeseries.plot(html=True)

        # True output len: 6564
        self.assertTrue(len(html_output)>6000)
        self.assertTrue(len(html_output)<7000)



import unittest
import os
import tempfile
from math import sin, cos
from propertime.utils import dt

from ..datastructures import TimePoint, DataTimeSlot, DataTimePoint, TimeSeries
from ..models.forecasters import ProphetForecaster, PeriodicAverageForecaster, ARIMAForecaster, AARIMAForecaster, LSTMForecaster
from ..exceptions import NonContiguityError
from ..storages import CSVFileStorage
from .. import TEST_DATASETS_PATH

# Setup logging
from .. import logger
logger.setup()

# Test data and temp path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'
TEMP_MODELS_DIR = tempfile.TemporaryDirectory().name


class TestForecasters(unittest.TestCase):

    def setUp(self):

        # Create a minute-resolution test DataTimeSlotSeries
        self.sine_minute_timeseries = TimeSeries()
        for i in range(1000):
            self.sine_minute_timeseries.append(DataTimeSlot(start=TimePoint(i*60), end=TimePoint((i+1)*60), data={'value':sin(i/10.0)}))

        # Create a day-resolution test DataTimeSlotSeries
        self.sine_day_timeseries = TimeSeries()
        for i in range(1000):
            step = 60 * 60 * 24
            self.sine_day_timeseries.append(DataTimeSlot(start=TimePoint(i*step), end=TimePoint((i+1)*step), data={'value':sin(i/10.0)}))

    @staticmethod
    def assertCompatible(actual, predicted, delta=0.3, item=None):
        if not abs(actual-predicted) < delta:
            raise AssertionError('Actual and predicted values are not compatible within the specified delta (actual="{}", predicted="{}", delta="{}", item="{}"'.format(actual, predicted, delta, item))


    def test_BaseForecasters(self):

        # TODO: This test is done using the Periodic Average Forecaster, make a mock

        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':1}, data_loss=0))
        timeseries.append(DataTimePoint(t=2,  data={'a':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=3,  data={'a':3}, data_loss=1))
        timeseries.append(DataTimePoint(t=4,  data={'a':4}, data_loss=0))

        forecaster = PeriodicAverageForecaster()
        forecaster.fitted = True
        forecaster.data['resolution'] = '1s'
        forecaster.data['data_labels'] = ['a']
        forecaster.data['window'] = 5

        # Not enough widow data for the predict, will raise
        with self.assertRaises(ValueError):
            forecaster.predict(timeseries)


    def test_PeriodicAverageForecaster(self):

        forecaster = PeriodicAverageForecaster()

        # Fit
        forecaster.fit(self.sine_minute_timeseries, periodicity=63)

        # Apply
        sine_minute_timeseries_with_forecast = forecaster.apply(self.sine_minute_timeseries, steps=3)
        self.assertEqual(len(sine_minute_timeseries_with_forecast), 1003)

        # Predict
        prediction = forecaster.predict(self.sine_minute_timeseries, steps=3)
        self.assertTrue(isinstance(prediction, list))
        self.assertEqual(len(prediction), 3)

        # Evaluate
        evaluation = forecaster.evaluate(self.sine_minute_timeseries[0:100], steps=1)
        self.assertEqual(forecaster.data['periodicities']['value'], 63)
        self.assertAlmostEqual(evaluation['value_RMSE'], 0.0873, places=2)
        self.assertAlmostEqual(evaluation['value_MAE'], 0.0797, places=2)

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = PeriodicAverageForecaster()
        with self.assertRaises(Exception):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        # TODO: do some actual testing.. not only that "it works"
        forecasted_timeseries  = forecaster.apply(timeseries)


    def test_PeriodicAverageForecaster_multivariate(self):
        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with multivariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))
        forecaster = PeriodicAverageForecaster()
        forecaster.fit(timeseries)

        # TODO: do some actual testing.. not only that "it works"
        timeseries_with_forecast = forecaster.apply(timeseries[0:970], steps=5)


    def test_PeriodicAverageForecaster_save_load(self):

        forecaster = PeriodicAverageForecaster()

        forecaster.fit(self.sine_minute_timeseries, periodicity=63)

        model_path = TEMP_MODELS_DIR + '/test_PA_model'

        forecaster.save(model_path)

        loaded_forecaster = PeriodicAverageForecaster.load(model_path)

        self.assertEqual(forecaster.data['offsets_averages'], loaded_forecaster.data['offsets_averages'])

        # TODO: do some actual testing.. not only that "it works"
        forecasted_timeseries = loaded_forecaster.apply(self.sine_minute_timeseries)


    def test_ProphetForecaster(self):

        try:
            import prophet
        except ImportError:
            print('Skipping Prophet tests as no prophet module installed')
            return

        forecaster = ProphetForecaster()

        forecaster.fit(self.sine_day_timeseries)
        self.assertEqual(len(self.sine_day_timeseries), 1000)

        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # Test the evaluate
        evalation_results = forecaster.evaluate(self.sine_day_timeseries.view(0,10))
        self.assertAlmostEqual(evalation_results['value_RMSE'], 0.80, places=2)
        self.assertAlmostEqual(evalation_results['value_MAE'], 0.77, places=2)

        evalation_results = forecaster.evaluate(self.sine_day_timeseries.view(0,1))
        self.assertAlmostEqual(evalation_results['value_RMSE'], 0.53, places=2) # For one sample they must be the same
        self.assertAlmostEqual(evalation_results['value_MAE'], 0.53, places=2) # For one sample they must be the same

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = ProphetForecaster()
        with self.assertRaises(Exception):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        # TODO: do some actual testing.. not only that "it works"
        forecasted_timeseries = forecaster.apply(timeseries)


    def test_ARIMAForecaster(self):

        try:
            import statsmodels
        except ImportError:
            print('Skipping ARIMA tests as no statsmodels module installed')
            return

        import warnings
        from statsmodels.tools.sm_exceptions import ConvergenceWarning
        warnings.simplefilter('ignore', ConvergenceWarning)

        # Basic ARIMA
        forecaster = ARIMAForecaster(p=1,d=1,q=0)

        forecaster.fit(self.sine_day_timeseries)
        self.assertEqual(len(self.sine_day_timeseries), 1000)

        # Cannot apply on a time series contiguous with the time series used for the fit
        with self.assertRaises(NonContiguityError):
            forecaster.apply(self.sine_day_timeseries[:-1], steps=3)

        # Can apply on a time series contiguous with the fit one
        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # # Cannot evaluate on a time series not contiguous with the time series used for the fit
        # with self.assertRaises(NonContiguityError):
        #     forecaster.evaluate(self.sine_day_timeseries)
        #
        # # Can evaluate on a time series contiguous with the time series used for the fit
        # forecaster = ARIMAForecaster(p=1,d=1,q=0)
        # forecaster.fit(self.sine_day_timeseries[0:800])
        # evaluation_results = forecaster.evaluate(self.sine_day_timeseries[800:1000])
        # self.assertAlmostEqual(evaluation_results['RMSE'], 2.71, places=2)
        # self.assertAlmostEqual(evaluation_results['MAE'], 2.52, places=2 )

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = ARIMAForecaster()
        with self.assertRaises(ValueError):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries)

        # TODO: do some actual testing.. not only that "it works"
        forecasted_timeseries  = forecaster.apply(timeseries)


    def test_AARIMAForecaster(self):

        try:
            import statsmodels
        except ImportError:
            print('Skipping AARIMA tests as no statsmodels module installed')
            return

        # Automatic ARIMA
        forecaster = AARIMAForecaster()

        forecaster.fit(self.sine_day_timeseries, max_p=2, max_d=1, max_q=2)
        self.assertEqual(len(self.sine_day_timeseries), 1000)

        # Cannot apply on a time series contiguous with the time series used for the fit
        with self.assertRaises(NonContiguityError):
            forecaster.apply(self.sine_day_timeseries[:-1], steps=3)

        # Can apply on a time series contiguous with the same item as the fit one
        sine_day_timeseries_with_forecast = forecaster.apply(self.sine_day_timeseries, steps=3)
        self.assertEqual(len(sine_day_timeseries_with_forecast), 1003)

        # # Cannot evaluate on a time series not contiguous with the time series used for the fit
        # with self.assertRaises(NonContiguityError):
        #     forecaster.evaluate(self.sine_day_timeseries)
        #
        # # Can evaluate on a time series contiguous with the time series used for the fit
        # forecaster = AARIMAForecaster()
        # forecaster.fit(self.sine_day_timeseries[0:800], max_p=2, max_d=1, max_q=2)
        # evaluation_results = forecaster.evaluate(self.sine_day_timeseries[800:1000])
        # self.assertTrue('RMSE' in evaluation_results)
        # self.assertTrue('MAE' in evaluation_results)
        # # Cannot test values, some random behavior which cannot be put under control is present somewhere
        # #self.assertAlmostEqual(evaluation_results['RMSE'], 0.7179428895746799)
        # #self.assertAlmostEqual(evaluation_results['MAE'], 0.6497934134525981)

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        forecaster = AARIMAForecaster()
        with self.assertRaises(ValueError):
            forecaster.fit(timeseries)

        timeseries = timeseries.resample(600)
        forecaster.fit(timeseries, max_p=2, max_d=1, max_q=2)

        # TODO: do some actual testing.. not only that "it works"
        forecasted_timeseries  = forecaster.apply(timeseries)


    def test_LSTMForecaster_univariate(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with univariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0)}))

        # Basic test
        forecaster = LSTMForecaster(window=10)
        forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)
        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] <0.2, evaluation_results['sin_MAE'])

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted)

        # Test using other parameters and features
        forecaster = LSTMForecaster(window=10, neurons=64, features=['values','diffs'])
        forecaster.fit(timeseries[0:970], epochs=10, reproducible=True, verbose=False)
        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] <0.2, evaluation_results['sin_MAE'])

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted, item=i)

        # Test for non-existent features
        with self.assertRaises(ValueError):
            LSTMForecaster(features=['values','not_existent_feature']).fit(timeseries)


    def test_LSTMForecaster_multivariate(self):
        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with multivariate time series as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))

        # Basic test
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries, epochs=10, reproducible=True)
        evaluation_results = forecaster.evaluate(timeseries[0:100], error_metrics=['MAE'], return_evaluation_series=True)

        self.assertTrue(evaluation_results['sin_MAE'] < 0.05, evaluation_results['sin_MAE'])
        self.assertTrue(evaluation_results['sin_MAE'] < 0.05, evaluation_results['sin_MAE'])

        timeseries_forecast = forecaster.apply(timeseries[0:970], steps=10)

        for i in range(10):
            actual = timeseries[970+i].data['sin']
            predicted = timeseries_forecast[970+i].data['sin']
            self.assertCompatible(actual, predicted)

        for i in range(10):
            actual = timeseries[970+i].data['cos']
            predicted = timeseries_forecast[970+i].data['cos']
            self.assertCompatible(actual, predicted)


    def test_LSTMForecaster_multivariate_with_targets_and_context(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster tests with multivariate time series with targets as no tensorflow module installed')
            return

        timeseries = TimeSeries()
        for i in range(980):
            timeseries.append(DataTimePoint(t=i*60, data={'sin': sin(i/10.0), 'cos': cos(i/10.0)}))

        # Target, no context
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries[0:978], epochs=10, target='cos', with_context=False, reproducible=True)

        with self.assertRaises(ValueError):
            forecaster.predict(timeseries[0:978], context_data={'sin': 'whatever'})

        predicted_data = forecaster.predict(timeseries)
        self.assertEqual(list(predicted_data.keys()), ['cos'])
        self.assertCompatible(timeseries[979].data['cos'], predicted_data['cos'])

        # Target and context
        forecaster = LSTMForecaster()
        forecaster.fit(timeseries[0:978], target='cos', with_context=True, reproducible=True)

        with self.assertRaises(ValueError):
            forecaster.predict(timeseries[0:978])

        predicted_data = forecaster.predict(timeseries[0:978], context_data={'sin': timeseries[979].data['cos']})
        self.assertEqual(list(predicted_data.keys()), ['cos'])
        self.assertCompatible(timeseries[979].data['cos'], predicted_data['cos'])

        # All other methods will raise as not suitable
        with self.assertRaises(ValueError):
            forecaster.forecast(timeseries)

        with self.assertRaises(ValueError):
            forecaster.apply(timeseries)




    def test_LSTMForecaster_cross_validation(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM forecaster cross validation tests as no tensorflow module installed')
            return

        temperature_timeseries = TimeSeries.from_csv(TEST_DATASETS_PATH + 'temperature_winter.csv').resample('1h')

        # Pretend there was no data loss at all
        for item in temperature_timeseries:
            item.data_indexes['data_loss'] = 0 
        forecaster = LSTMForecaster(window=12, neurons=64, features=['values', 'diffs', 'hours'])
        cross_validation_results = forecaster.cross_validate(temperature_timeseries[0:100], rounds=3, fit_reproducible=True)

        self.assertTrue(0.3 < cross_validation_results['temperature_RMSE_avg'] < 0.6) #0.4926859886937329
        self.assertTrue(0.2 < cross_validation_results['temperature_MAE_avg'] < 0.5) #0.39961679741401374


    def test_LSTMForecaster_save_load(self):

        try:
            import tensorflow
        except ImportError:
            print('Skipping LSTM save/load forecaster tests as no tensorflow module installed')
            return

        # Create a minute-resolution test DataTimeSlotSeries
        sine_minute_timeseries = TimeSeries()
        for i in range(10):
            sine_minute_timeseries.append(DataTimeSlot(start=TimePoint(i*60), end=TimePoint((i+1)*60), data={'sin':sin(i/10.0), 'cos':cos(i/10.0)}))

        # Fit a forecaster and make a prediction
        forecaster = LSTMForecaster()
        forecaster.fit(sine_minute_timeseries, reproducible=True)
        predicted_data = forecaster.predict(sine_minute_timeseries)

        # Set model path
        model_path = TEMP_MODELS_DIR+ '/test_LSTM_model'

        # Save
        forecaster.save(model_path)

        # Load
        forecaster_loaded = LSTMForecaster.load(model_path)

        # Predict from the loaded model
        predicted_data_loaded = forecaster_loaded.predict(sine_minute_timeseries)

        # Check predictions are the same
        self.assertEqual(predicted_data, predicted_data_loaded)


    def test_LinearRegressionForecaster(self):

        # Create a hour-resolution test DataTimeSlotSeries
        sine_hour_timeseries = TimeSeries()
        for i in range(1000):
            sine_hour_timeseries.append(DataTimePoint(t=i*3600, data={'value':sin(i/10.0)}))

        from ..models import LinearRegressionForecaster
        forecaster = LinearRegressionForecaster(features=['values','hours'])

        # Fit
        forecaster.fit(sine_hour_timeseries)

        # Predict and check
        self.assertAlmostEqual(forecaster.apply(sine_hour_timeseries)[-1].data['value'], -0.5063, places=3)

        # Save
        forecaster.save(TEMP_MODELS_DIR+'/test_lr_model')

        # Load
        loaded_forecaster = LinearRegressionForecaster.load(TEMP_MODELS_DIR+'/test_lr_model')

        # Re-predict and re-check
        self.assertAlmostEqual(loaded_forecaster.apply(sine_hour_timeseries)[-1].data['value'], -0.5063, places=3)




# -*- coding: utf-8 -*-
"""Anomaly detection models."""

from copy import deepcopy
from ..utils import _Gaussian, rescale
from .forecasters import Forecaster, PeriodicAverageForecaster, LSTMForecaster, LinearRegressionForecaster
from .reconstructors import Reconstructor, PeriodicAverageReconstructor
from .base import Model
from math import log10, prod
from fitter import Fitter, get_common_distributions, get_distributions
from ..utils import DistributionFunction
from statistics import stdev
import fitter as fitter_library
from ..datastructures import TimeSeries, DataTimePoint

# Setup logging
import logging
logger = logging.getLogger(__name__)

fitter_library.fitter.logger = logging.getLogger('fitter')
fitter_library.fitter.logger.setLevel(level=logging.CRITICAL)

# Suppress TensorFlow warnings as default behavior
try:
    import tensorflow as tf
    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
except (ImportError,AttributeError):
    pass


#===================================
#  Generic Anomaly Detector
#===================================

class AnomalyDetector(Model):
    """A generic anomaly detection model."""

    def predict(self, series, *args, **kwargs):
        """Disabled. Anomaly detectors can be used only with the ``apply()`` method."""
        raise NotImplementedError('Anomaly detectors can be used only with the apply() method') from None

    def evaluate(self, series, *args, **kwargs):
        """Disabled. Anomaly detectors cannot be evaluated yet."""
        raise NotImplementedError('Anomaly detectors cannot be evaluated yet.') from None

    def cross_validate(self, series, *args, **kwargs):
        """Disabled. Anomaly detectors cannot be evaluated yet."""
        raise NotImplementedError('Anomaly detectors cannot be evaluated yet.') from None

    @staticmethod
    def mark_events(timeseries, index_treshold=1.0, min_persistence=2, max_gap=2, replace_index=False, inplace=False):
        """Mark ensembles of anomalous data points as single anomalous events, with some tolerance.

        Args:
            index_treshold(float): the anomaly index above which to consider a data point anomalous.
            min_persistence(int): the minimum persistence of an event, in terms of consecutive data points.
            max_gap(int): the maximum gap within a signle event of data points below the index_treshold.
            replace_index(bool): if to replace the existent ``anomaly`` index instead of adding a new ``anomaly_event`` one.
        """

        if inplace:
            event_timeseries = timeseries
        else:
            event_timeseries = timeseries.duplicate()
        event_start = None

        for i, item in enumerate(event_timeseries):

            # Set no event by default
            item.data_indexes['anomaly_event'] = 0

            # Detect new (potential) event start
            if item.data_indexes['anomaly'] >= index_treshold:
                if event_start is None:
                    logger.debug('Starting to suspect event @ element #%s',i)
                    event_start = i

            # Detect event end
            if event_start is not None:
                if item.data_indexes['anomaly'] < index_treshold:
                    logger.debug('Evaluating element #%s to trigger an event end', i)

                    # Are we still within the max gap? If so, don't close the (potential) event yet
                    from_i = (i - max_gap) if (i - max_gap) >= 0 else 0
                    to_i = i
                    event_ended = True
                    logger.debug('  checking elements from #%s to #%s', from_i, to_i)
                    for j in range(from_i, to_i+1):
                        if event_timeseries[j].data_indexes['anomaly'] >= index_treshold:
                            event_ended = False
                    if event_ended:
                        logger.debug('  event was ended @ element #%s', i - max_gap - 1)
                        event_persistence =  (i - max_gap) - event_start
                        if event_persistence < min_persistence:
                            logger.debug('  discarding as below minimum persistence')
                        else:
                            # Mark it
                            for j in range(event_start, i - max_gap):
                                event_timeseries[j].data_indexes['anomaly_event'] = 1

                        # Reset the event_start support var
                        event_start = None
                else:
                    logger.debug('Element #%s is part of the event', i)

        # Handle series last event, if any
        if event_start is not None:

            # When was the last event over?
            event_end = len(event_timeseries)
            for i in range(len(event_timeseries)-1, event_start, -1):
                if  event_timeseries[i].data_indexes['anomaly'] >= index_treshold:
                    event_end = i
                    break
            logger.debug('  event was ended @ element #%s', event_end)
            event_persistence =  event_end - event_start
            if event_persistence < min_persistence:
                logger.debug('  discarding as below minimum persistence')
            else:
                # Mark it
                for j in range(event_start, event_end+1):
                    event_timeseries[j].data_indexes['anomaly_event'] = 1

        if replace_index:
            for item in event_timeseries:
                item.data_indexes['anomaly'] = item.data_indexes.pop('anomaly_event') 

        return  event_timeseries


#===================================
#  Model-based Anomaly Detector
#===================================

class ModelBasedAnomalyDetector(AnomalyDetector):
    """An anomaly detection model based on another model. See the ``model_class`` property for which model is being used.

    In short, for each element of the series where the anomaly detector is applied, the model is asked to make a prediction.
    The predicted and actual values are then compared, and accordingly to the model error distribution, an anomaly index
    is computed. The anomaly index is proportional to the likelihood of each element to be anomalous, and ranges form 0 to 1.

    More in detail, in the fit phase the model set in the ``model_class`` is first fitted and then evaluated using the error
    metric set in the ``error_metric`` argument, which defaults to PE (Percentage Error). All of the error values recorded in
    the evaluation are then used to build the error distribution (one for each data label of the series), which defaults to a
    generalized normal distribution. If the p-value is low (below 0.05), a warning is issued.

    When the anomaly detector is then applied, the anomaly index is computed based on two boundaries of such error distribution,
    which are set in the ``index_bounds`` argument of the ``apply()`` method. Error values below the lower bound will map to an
    anomaly index of zero (no anomaly suspected at all), between the lower and the upper bound will map to anomaly indexes between
    zero and one (some degree of anomaly is suspected) and above the upper bound will map to an anomaly index of one (certainly
    an anomaly).

    How to set such boundaries depends on many factors, the most important of which being whether the anomaly detector is used in
    usupervised or semi-supervised mode. In unsupervised mode, a reasonable choice would be to set the lower bound to the average
    error, and the upper bound to the maximum error. In semi-supervised mode (which is way more powerful since a notion of normality
    is implicitly given in the fit data), a good choice would be to set the lower bound to the maximum error (which still belongs to
    the "normal" behavior) and the upper bound to a very high error (e.g. a billion times the probability of an observation to be
    adherent with the model). More details about how to set the boundaries are provided in the ``apply()`` method documentation.

    Args:
        model_class(Model): the model to be used for anomaly detection, if not already set.
        index_window(int): the (rolling) window length to be used when computing the anomaly index. Defaults to 1.
    """

    @property
    def model_class(self):
        try:
            return self._model_class
        except AttributeError:
            raise NotImplementedError('No model class set for this anomaly detector')

    def __init__(self, model_class=None, index_window=1, *args, **kwargs):

        # Handle the model_class
        try:
            self.model_class
        except NotImplementedError:
            if not model_class:
                raise ValueError('The model_class is not set in the anomaly detector nor given in the init') from None
            else:
                self._model_class = model_class
        else:
            if model_class:
                raise ValueError('The model_class was given in the init but it is already set in the anomaly detector')

        self.predictive_model_args = args
        self.predictive_model_kwargs = kwargs
        self.index_window = index_window

        # Call parent init
        super(ModelBasedAnomalyDetector, self).__init__()
 
    @classmethod
    def load(cls, path):

        # Load the anomaly detection model
        model = super().load(path)

        # ..and load the inner model(s) as nested model
        if model.data['with_context']:
            model.models = {}
            for data_label in model.data['data_labels']:
                model.models[data_label] = model.model_class.load('{}/{}_model'.format(path, data_label))
        else:
            model.model= model.model_class.load('{}/model'.format(path))

        return model

    def save(self, path):

        # Save the anomaly detection model
        super(ModelBasedAnomalyDetector, self).save(path)

        # ..and save the inner model(s) as nested model
        if self.data['with_context']:
            for data_label in self.models:
                self.models[data_label].save('{}/{}_model'.format(path, data_label))
        else:
            self.model.save('{}/model'.format(path))

    def _get_actual_value(self, series, i, data_label):
        actual = series[i].data[data_label]
        return actual

    def _get_predicted_value(self, series, i, data_label, with_context):

            # Call model predict logic and compare with the actual data
            if issubclass(self.model_class, Reconstructor):
                prediction = self.model.predict(series, from_i=i, to_i=i)

            elif issubclass(self.model_class, Forecaster):

                if with_context:
                    prediction = self.models[data_label].predict(series.view(from_i=0, to_i=i), steps=1, context_data=series[i].data)
                else:
                    prediction = self.model.predict(series.view(from_i=0, to_i=i), steps=1)

            else:
                raise TypeError('Don\'t know how to handle predictive model class "{}"'.format(self.model_class.__name__))

            # Handle list of dicts or dict of lists (of which we have only one value here)
            #{'value': [0.2019341593004146, 0.29462641146884005]}

            if isinstance(prediction, list):
                predicted = prediction[0][data_label]
            elif isinstance(prediction, dict):
                if isinstance(prediction[data_label], list):
                    predicted = prediction[data_label][0]
                else:
                    predicted = prediction[data_label]
            else:
                raise TypeError('Don\'t know how to handle a prediction with of type "{}"'.format(prediction.__class__.__name__))

            return predicted

    @AnomalyDetector.fit_method
    def fit(self, series, with_context=False, error_metric='PE', error_distribution='gennorm', store_errors=True, verbose=False, summary=False, **kwargs):
        """Fit the anomaly detection model on a series.

        Args:
            with_context(bool): if to use context for multivariate time series or not. Defaults to ``False``.
            error_metric(str): the error metric to be used for evaluating the model and to build the error distribution for the anomaly index.
                               Supported values are: ``E``, ``AE``, ``PE`` and ``APE``. Defaults to ``PE``.
            error_distribution(str): if to use a specific error distribution or find it automatically (``error_distribution='auto'``).
                                     Defaults to ``gennorm``, a generalized normal distribution.
            store_errors(float): if to store the prediction errors (together with actual and predicted values) internally for further analysis. Access
                                 them with ``model.data['prediction_errors']``, ``model.data['actual_values']`` and ``model.data['predicted_values']``.
            verbose(bool): if to print the fit progress (one dot = 10% done).
            summary(bool): if to display a summary on the error distribution fitting or selection.
        """

        # Handle the error distribution(s)
        if error_distribution == 'auto':
            error_distributions = kwargs.pop('error_distributions', fitter_library.fitter.get_common_distributions() +['gennorm'])
        else:
            error_distributions = [error_distribution]

        # Fit the predictive model(s)
        if with_context:
            if not len(series.data_labels()) > 1:
                raise ValueError('Anomaly detection with partial predictions on univariate series does not make sense')

            # Fit separate models, one for each data label
            self.models = {}
            self.data['model_ids'] = {}
            for data_label in series.data_labels():
                if verbose:
                    print('Fitting for "{}":'.format(data_label))
                logger.debug('Fitting for "%s"...', data_label)

                # Initialize the internal model for this label
                self.models[data_label] = self.model_class(*self.predictive_model_args, **self.predictive_model_kwargs)

                # Set the id of the internal model in the data
                self.data['model_ids'][data_label] = self.models[data_label].data['id']

                # Fit it
                self.models[data_label].fit(series, **kwargs, target=data_label, with_context=True, verbose=verbose)

                # Set the model window if not already done:
                if 'model_window' not in self.data:
                    self.data['model_window'] = self.models[data_label].window

        else:
            # Initialize the internal model
            self.model = self.model_class(*self.predictive_model_args, **self.predictive_model_kwargs)

            # Set the id of the internal model in the data
            self.data['model_id'] = self.model.data['id']

            # Fit it
            self.model.fit(series, **kwargs, verbose=verbose)

            # Set the model window
            self.data['model_window'] = self.model.window

        if verbose:
            print('Predictive model(s) fitted, now evaluating')
        logger.info('Predictive model(s) fitted, now evaluating...')

        # Store if to use context or not
        self.data['with_context'] = with_context

        # Check and store error metric
        if error_metric not in ['E', 'AE', 'PE', 'APE', 'SLE']:
            raise ValueError('Unknown error metric "{}"'.format(error_metric))
        self.data['error_metric'] = error_metric

        # Initialize internal dictionaries
        if store_errors:
            self.data['actual_values'] = {}
            self.data['predicted_values'] = {}
            self.data['prediction_errors'] = {}
        self.data['stdevs'] = {} 
        self.data['error_distributions'] = {}
        self.data['error_distributions_params'] = {}
        self.data['error_distributions_stats'] = {}

        # Evaluate the predictive for one step ahead and get the forecasting errors
        actual_values = {}
        predicted_values = {}
        prediction_errors = {}
        cumulative_prediction_errors = {}

        progress_step = len(series)/10

        for data_label in series.data_labels():

            actual_values[data_label] = []
            predicted_values[data_label] = []
            prediction_errors[data_label] = []
            cumulative_prediction_errors[data_label] = []

            if verbose:
                print('Computing actual vs predicted for "{}": '.format(data_label), end='')
            logger.info('Computing actual vs predicted for "{}"...'.format(data_label))

            for i, _ in enumerate(series):
                if verbose:
                    if int(i%progress_step) == 0:
                        print('.', end='')

                # Before the window
                if i <=  self.data['model_window']:
                    continue

                # After the window (if using a reconstructor)
                if issubclass(self.model_class, Reconstructor):
                    if i > len(series)-self.data['model_window']-1:
                        break

                # Predict & append the error
                #logger.debug('Predicting and computing the difference (i=%s)', i)
                actual = self._get_actual_value(series, i, data_label)
                predicted = self._get_predicted_value(series, i, data_label, with_context)

                if store_errors:
                    actual_values[data_label].append(actual)
                    predicted_values[data_label].append(predicted)

                if error_metric == 'E':
                    prediction_error = actual-predicted
                elif error_metric == 'AE':
                    prediction_error = abs(actual-predicted)
                elif error_metric == 'PE':
                    prediction_error = (actual-predicted)/actual
                elif error_metric == 'APE':
                    prediction_error = abs((actual-predicted)/actual)
                else:
                    raise ValueError('Unknown error metric "{}"'.format(self.data['error_metric']))

                prediction_errors[data_label].append(prediction_error)
                if error_metric in ['AE', 'APE']:
                    prediction_errors[data_label].append(-prediction_error)

                # Handle the index window if any
                if (self.index_window > 1) and (i >= self.data['model_window'] + self.index_window):
                    if error_metric in ['AE', 'APE']:
                        raise NotImplementedError('Error metrics bases on absolute values are not yet supported when using a window for the anomaly index')
                        cumulative_error = 0
                        for j, item in enumerate(prediction_errors[data_label][-self.index_window*2:]):
                            if j % 2 == 0:
                                cumulative_error += item
                        cumulative_prediction_errors[data_label].append(cumulative_error) # Likely a bi-modal distribution
                        #cumulative_prediction_errors[data_label].append(-cumulative_error) # Likely a gamma distribution
                    else:
                        cumulative_error = sum(prediction_errors[data_label][-self.index_window:])
                        cumulative_prediction_errors[data_label].append(cumulative_error)

            if verbose:
                print('')

        if self.index_window > 1:
            prediction_errors = cumulative_prediction_errors

        if verbose:
            print('Model(s) evaluated, now computing the error distribution(s)')
        logger.info('Model(s) evaluated, now computing the error distribution(s)...')

        for data_label in series.data_labels():
            #if verbose:
            #    print('Selecting error distribution for "{}"'.format(data_label))
            #logger.debug('Selecting error distribution for "%s"', data_label))

            # Fit the distributions and select the best one
            fitter = fitter_library.fitter.Fitter(prediction_errors[data_label], distributions=error_distributions)
            fitter.fit(progress=False)

            if summary:
                # Warning: the summary() function will also generate a plot
                print(fitter.summary())

            best_error_distribution = list(fitter.get_best().keys())[0]
            best_error_distribution_stats = fitter.summary(plot=False).transpose().to_dict()[best_error_distribution]
            error_distribution_params = fitter.get_best()[best_error_distribution]

            if best_error_distribution_stats['ks_pvalue'] < 0.05:
                logger.warning('The error distribution for "{}" ({}) p-value is low: {}.'.format(data_label, best_error_distribution, best_error_distribution_stats['ks_pvalue']))

            self.data['error_distributions'][data_label] = best_error_distribution
            self.data['error_distributions_params'][data_label] = error_distribution_params
            self.data['error_distributions_stats'][data_label] = best_error_distribution_stats

            self.data['stdevs'][data_label] = stdev(prediction_errors[data_label])

        if store_errors:
            self.data['actual_values'] = actual_values
            self.data['predicted_values'] = predicted_values
            self.data['prediction_errors'] = prediction_errors

        self.data['index_window'] = self.index_window

        logger.info('Anomaly detector fitted')

    def inspect(self, plot=True, plot_x_min='auto', plot_x_max='auto', series=False):
        '''Inspect the model and plot the error distribution'''

        for data_label in self.data['error_distributions']:

            abs_prediction_errors = [abs(prediction_error) for prediction_error in self.data['prediction_errors'][data_label]]

            print('\nDetails for: "{}"'.format(data_label))

            if self.data['index_window']  >1:
                print('Predictive model avg error (abs, {} items): {}'.format(self.data['index_window'], sum(abs_prediction_errors)/len(abs_prediction_errors)))
                print('Predictive model min error (abs, {} items): {}'.format(self.data['index_window'], min(abs_prediction_errors), self.data['index_window']))
                print('Predictive model max error (abs, {} items): {}'.format(self.data['index_window'], max(abs_prediction_errors), self.data['index_window']))
            else:
                print('Predictive model avg error (abs): {}'.format(sum(abs_prediction_errors)/len(abs_prediction_errors)))
                print('Predictive model min error (abs): {}'.format(min(abs_prediction_errors)))
                print('Predictive model max error (abs): {}'.format(max(abs_prediction_errors)))

            print('Error distribution: {}'.format(self.data['error_distributions'][data_label]))
            print('Error distribution params: {}'.format(self.data['error_distributions_params'][data_label]))
            print('Error distribution stats: {}'.format(self.data['error_distributions_stats'][data_label]))

            if plot:

                if plot_x_min == 'auto':
                    plot_x_min = min(self.data['prediction_errors'][data_label])
                if plot_x_max == 'auto':
                    plot_x_max = max(self.data['prediction_errors'][data_label])

                # Instantiate the error distribution function
                distribution_function = DistributionFunction(self.data['error_distributions'][data_label],
                                                             self.data['error_distributions_params'][data_label])

                # Get the error distribution function plot
                plt = distribution_function.plot(show=False, x_min=plot_x_min, x_max=plot_x_max)

                # Add the histogram to the plot
                plt.hist(self.data['prediction_errors'][data_label], bins=100, density=True, alpha=1, color='steelblue')

                # Override title
                #plt.title('Error distribution: {}'.format(self.data['error_distribution']))

                # Show the plot
                plt.show()

            if series:
                if not 'prediction_errors' in self.data:
                    raise ValueError('Cannot inspect fit series if store_errors was not set.')
                fit_series = TimeSeries()
                first_data_label = list(self.data['actual_values'].keys())[0]
                for i in range(len(self.data['actual_values'][first_data_label])):
                    data = {}
                    for data_label in self.data['actual_values']:
                        data[data_label] = self.data['actual_values'][data_label][i]
                        data[data_label+'_pred'] = self.data['predicted_values'][data_label][i]
                        data[data_label+'_err'] = self.data['prediction_errors'][data_label][i]
                    fit_series.append(DataTimePoint(t=i, data=data))
                fit_series.plot()

    def preprocess(self, series, inplace=False, verbose=False):
        """Pre-process a time series for this anomaly detector so that multiple apply() calls are much faster"""
        if not inplace:
            series = series.duplicate()

        # Start pre-processing
        series.predictions = {}
        progress_step = len(series)/10
        if verbose:
            print('Pre-computing model predictions: ', end='')
        for i, item in enumerate(series):

            if verbose:
                if int(i%progress_step) == 0:
                    print('.', end='')

            # Before the window
            if i <=  self.data['model_window']:
                continue

            # After the window (if using a reconstructor)
            if issubclass(self.model_class, Reconstructor):
                if i >  len(series)-self.data['model_window']-1:
                    break

            ith_predictions = {} 
            for data_label in series.data_labels():
                ith_predictions[data_label] = self._get_predicted_value(series, i, data_label, self.data['with_context'])

            series.predictions[i] = ith_predictions

        series.preprocessed_by = self.id
        return series

    @Model.apply_method
    def apply(self, series, index_bounds=['avg_err','max_err'], index_type='log', multivariate_index_strategy='max',
              data_loss_threshold=1.0, details=False, verbose=False):

        """Apply the anomaly detection model on a series.

        Args:
            index_bounds(tuple): the lower an upper bounds, in terms of the prediction error, from which the anomaly index is
                                 computed. Below the lower limit no anomaly will be assumed (``anomlay_index=0``), while errors
                                 above the upper will always be considered as anomalous (``anomlay_index=1``). In the middle, the
                                 anomaly index will range between 0 and 1 following the model error distribution. Defaults values
                                 are ``('avg_err', 'max_err')``, which are reasonable values for unsupervised anomaly detection.
                                 Other supported values are: ``x_sigma``, where x is a standard deviation multiplier;
                                 ``adherence/x``, where x is a divider for the model adherence probability; and any other
                                 numerical value in terms of prediction error.
            index_type(str, callable): if to use a logarithmic anomaly index ("log", the default value) which compresses
                                       the index range so that bigger anomalies stand out more than smaller ones, or if to
                                       use a linear one ("lin"). Can also support a custom anomaly index as a callable,
                                       in which case the form must be ``f(x, y, x_start, x_end, y_start, y_end)`` where x
                                       is the model error, y its value on the distribution curve, and x_start/x_end together
                                       with y_start/y_end the respective x and y range start and end values, based on the
                                       range set by the ``index_bounds`` argument.
            multivariate_index_strategy(str, callable): the strategy to use when computing the overall anomaly index for multivariate
                                                        time series items. Possible choices are "max" to use the maximum one, "avg"
                                                        for the mean and "min" for the minimum; or a callable taking as input the
                                                        list of the anomaly indexes for each data label. Defaults to "max".
            data_loss_treshold(float): if the data loss is equal or greater than this threshold value, then the anomaly detection is not applied.
            details(bool, list): if to add details to the time series as the predicted value, the error and the
                                 model adherence probability. If set to ``True``, it adds all of them, if instead using
                                 a list only selected details can be added: ``pred`` for the predicted values, ``err`` for
                                 the errors, and ``adh`` for the model adherence probability.
            verbose(bool): if to print the apply progress (one dot = 10% done).
        """
        # Initialize the result time series
        result_series = series.__class__()

        # Initialize the error distribution function
        error_distribution_functions = {}
        for data_label in self.data['error_distributions']:
            error_distribution_functions[data_label] = DistributionFunction(self.data['error_distributions'][data_label],
                                                                            self.data['error_distributions_params'][data_label])

        # Check error metric
        if self.data['error_metric'] in ['E', 'AE', 'PE', 'APE']:
            pass
        elif callable(self.data['error_metric']):
            # If the model is loaded, it will never get here
            pass
        else:
            raise ValueError('Unknown error metric "{}"'.format(self.data['error']))

        # Set anomaly index boundaries
        x_starts = {}
        x_ends = {}
        y_starts = {}
        y_ends = {}
        log_10_y_starts = {}
        log_10_y_ends = {}
        y_maxes = {}
        rolling_prediction_errors = {}

        for data_label in series.data_labels():
            abs_prediction_errors = [abs(prediction_error) for prediction_error in self.data['prediction_errors'][data_label]]
            y_maxes[data_label] = error_distribution_functions[data_label](self.data['error_distributions_params'][data_label]['loc'])
            this_index_bounds = index_bounds[:]

            for i in range(2):
                if isinstance(this_index_bounds[i], str):
                    if this_index_bounds[i] == 'max_err':
                        this_index_bounds[i] = max(abs_prediction_errors)
                    elif this_index_bounds[i] == 'avg_err':
                        this_index_bounds[i] = sum(abs_prediction_errors)/len(abs_prediction_errors)
                    elif this_index_bounds[i].endswith('_sigma'):
                        this_index_bounds[i] = float(this_index_bounds[i].replace('_sigma',''))*self.data['stdevs'][data_label]
                    elif this_index_bounds[i].endswith('sig'):
                        this_index_bounds[i] = float(this_index_bounds[i].replace('sig',''))*self.data['stdevs'][data_label]
                    elif this_index_bounds[i].startswith('adherence/'):
                        factor = float(this_index_bounds[i].split('/')[1])
                        this_index_bounds[i] = error_distribution_functions[data_label].find_x(y_maxes[data_label]/factor)
                    else:
                        raise ValueError('Unknown index start or end value "{}"'.format(this_index_bounds[i]))

            x_starts[data_label] = this_index_bounds[0]
            x_ends[data_label] = this_index_bounds[1]

            # Compute error distribution function values for index start/end
            y_starts[data_label] = error_distribution_functions[data_label](x_starts[data_label])
            y_ends[data_label] = error_distribution_functions[data_label](x_ends[data_label])

            if index_type == 'log':
                try:
                    log_10_y_starts[data_label] = log10(y_starts[data_label] )
                except ValueError as e:
                    if str(e) == 'math domain error':
                        raise ValueError('Got a math domain error in computing the anomaly index start boundary for label "{}". This is likely due to extreme values and/or an error distribution badly approximating the real error distribution. Try changing the anomaly index boundaries or find a better error distribution.'.format(data_label)) from None
                    else:
                        raise
                try:
                    log_10_y_ends[data_label] = log10(y_ends[data_label] )
                except ValueError as e:
                    if str(e) == 'math domain error':
                        raise ValueError('Got a math domain error in computing the anomaly index end boundary for label "{}". This is likely due to extreme values and/or an error distribution badly approximating the real error distribution. Try changing the anomaly index boundaries or find a better error distribution.'.format(data_label)) from None
                    else:
                        raise

            # Prepare for the rolling anomaly index
            rolling_prediction_errors[data_label] = []

        # Is this series pre-processed?
        try:
            preprocessed = True if series.preprocessed_by == self.id else False
        except AttributeError:
            preprocessed = False

        # Start processing
        progress_step = len(series)/10
        if verbose:
            print('Applying the anomaly detector: ', end='')
        for i, item in enumerate(series):

            if verbose:
                if int(i%progress_step) == 0:
                    print('.', end='')

            # Before the window
            if i <=  self.data['model_window']:
                continue

            # After the window (if using a reconstructor)
            if issubclass(self.model_class, Reconstructor):
                if i >  len(series)-self.data['model_window']-1:
                    break

            # Check if we are allowed to compute the anomaly index given the data loss threshold, or if we have to abort
            abort=False
            if series[i].data_loss is not None and series[i].data_loss >= data_loss_threshold:
                abort=True
            else:
                for j in range(self.data['model_window']):
                    if series[i-j-1].data_loss is not None and series[i-j-1].data_loss >= data_loss_threshold:
                        abort=True
                        break
                if not abort:
                    if issubclass(self.model_class, Reconstructor):
                        if series[i-j-1].data_loss is not None and series[i+j+1].data_loss >= data_loss_threshold:
                            abort=True
                            break
            if abort:
                # Just append the item as-is and continue
                result_series.append(item)
                continue

            # Start computing the anomaly index
            item_anomaly_indexes = []

            for data_label in series.data_labels():

                # Shortcuts
                error_distribution_function = error_distribution_functions[data_label]
                x_start = x_starts[data_label]
                x_end = x_ends[data_label]
                y_start = y_starts[data_label]
                y_end = y_ends[data_label]
                if index_type == 'log':
                    log_10_y_start = log_10_y_starts[data_label]
                    log_10_y_end = log_10_y_ends[data_label]
                y_max = y_maxes[data_label]

                # Duplicate this series item
                item = deepcopy(item)

                # Compute the prediction error index
                if preprocessed:
                    actual = self._get_actual_value(series, i, data_label)
                    predicted = series.predictions[i][data_label]
                else:
                    actual = self._get_actual_value(series, i, data_label)
                    predicted = self._get_predicted_value(series, i, data_label, self.data['with_context'])

                if self.data['error_metric'] == 'E':
                    prediction_error = actual-predicted
                elif self.data['error_metric'] == 'AE':
                    prediction_error = abs(actual-predicted)
                elif self.data['error_metric'] == 'PE':
                    prediction_error = (actual-predicted)/actual
                elif self.data['error_metric'] == 'APE':
                    prediction_error = abs((actual-predicted)/actual)
                else:
                    raise ValueError('Unknown error type "{}"'.format(self.data['error_metric']))

                if self.data['error_metric']  in ['AE', 'APE']:
                    prediction_error = abs(prediction_error)

                # Handle the rolling window for the index if any
                if self.data['index_window'] > 1:
                    punctual_prediction_error = prediction_error
                    rolling_prediction_errors[data_label].append(prediction_error)
                    if (i >= self.data['model_window'] + self.data['index_window'] ):

                        # Compute the cumulative prediction errors on the rolling errors
                        prediction_error = sum(rolling_prediction_errors[data_label])

                        # Move the window
                        rolling_prediction_errors[data_label] = rolling_prediction_errors[data_label][1:]

                    else:
                        continue

                # Reverse values on the left side of the error distribution to simplify the math below
                distribution_loc= self.data['error_distributions_params'][data_label]['loc']
                if prediction_error < distribution_loc:
                    prediction_error =  distribution_loc + (distribution_loc - prediction_error)

                # Compute the anomaly index in the given range (which defaults to 0, max_err)
                # Below the start it means anomaly (0), above always anomaly (1). In the middle it
                # follows the error distribution, and it is rescaled between 0 and 1.

                x = prediction_error
                y = error_distribution_function(x)

                if x <= x_start:
                    anomaly_index = 0
                elif x >= x_end:
                    anomaly_index = 1
                else:
                    if index_type == 'lin':
                        anomaly_index = 1 - ((y-y_end)/(y_start-y_end))
                    elif index_type == 'log':
                        anomaly_index = (log10(y) - log_10_y_start) / (log_10_y_end - log_10_y_start)
                    else:
                        if callable(index_type):
                            anomaly_index = index_type(x, y, x_start, x_end, y_start, y_end)
                        else:
                            raise ValueError('Unknown index type "{}"'.format(index_type))

                # Add this anomaly index for this data label to the list of the item anomaly indexes
                item_anomaly_indexes.append(anomaly_index)

                # Add details?
                if details:
                    if self.data['index_window']>1:
                        if isinstance(details, list):
                            if 'pred' in details:
                                item.data['{}_pred'.format(data_label)] = predicted
                            if 'err' in details:
                                item.data['{}_err_cum'.format(data_label)] = prediction_error
                            if 'adh' in details:
                                item.data['{}_adh_cum)'.format(data_label)] = y/y_max
                        elif isinstance(details, bool):
                            item.data['{}_pred'.format(data_label)] = predicted
                            item.data['{}_err_cum'.format(data_label)] = prediction_error
                            item.data['{}_err'.format(data_label)] = punctual_prediction_error
                            item.data['{}_adh_cum'.format(data_label)] = y/y_max
                        else:
                            raise TypeError('The "details" argument accepts only True/False or a list containing what details to add, as strings.')
                    else:
                        if isinstance(details, list):
                            if 'pred' in details:
                                item.data['{}_pred'.format(data_label)] = predicted
                            if 'err' in details:
                                item.data['{}_err'.format(data_label)] = prediction_error
                            if 'adh' in details:
                                item.data['{}_adh)'.format(data_label)] = y/y_max
                        elif isinstance(details, bool):
                            item.data['{}_pred'.format(data_label)] = predicted
                            item.data['{}_err'.format(data_label)] = prediction_error
                            item.data['{}_adh'.format(data_label)] = y/y_max
                        else:
                            raise TypeError('The "details" argument accepts only True/False or a list containing what details to add, as strings.')

            # Set anomaly index & append
            if item_anomaly_indexes:
                if len(item_anomaly_indexes) == 1:
                    item.data_indexes['anomaly'] = item_anomaly_indexes[0]
                else:
                    if multivariate_index_strategy == 'min':
                        item.data_indexes['anomaly'] = min(item_anomaly_indexes)
                    elif multivariate_index_strategy == 'max':
                        item.data_indexes['anomaly'] = max(item_anomaly_indexes)
                    elif multivariate_index_strategy == 'avg':
                        item.data_indexes['anomaly'] = sum(item_anomaly_indexes)/len(item_anomaly_indexes)
                    else:
                        item.data_indexes['anomaly'] = multivariate_index_strategy(item_anomaly_indexes)

                # Append
                result_series.append(item)

        if verbose:
            print('')

        return result_series

    @property
    def error_distribution_function(self):
        distribution_function = DistributionFunction(self.data['error_distribution'],
                                                     self.data['error_distribution_params'])
        return distribution_function


#===================================
#   Periodic Average Forecaster
#   Predictive Anomaly Detector
#===================================

class PeriodicAverageAnomalyDetector(ModelBasedAnomalyDetector):
    """An anomaly detection model based on a periodic average forecaster."""

    model_class = PeriodicAverageForecaster


#===================================
#  Periodic Average Reconstructor
#   Predictive Anomaly Detector
#===================================

class PeriodicAverageReconstructorAnomalyDetector(ModelBasedAnomalyDetector):
    """An anomaly detection model based on a periodic average reconstructor."""

    model_class = PeriodicAverageReconstructor


#===================================
#  LSTM Anomaly Detector
#===================================

class LSTMAnomalyDetector(ModelBasedAnomalyDetector):
    """An anomaly detection model based on a LSTM neural network forecaster."""

    model_class = LSTMForecaster

#===================================
# Linear Regression Anomaly Detector
#===================================

class LinearRegressionAnomalyDetector(ModelBasedAnomalyDetector):
    """An anomaly detection model based on a linear regression forecaster."""

    model_class = LinearRegressionForecaster




import unittest
import os
import tempfile
from propertime.utils import dt

from ..datastructures import TimePoint, DataTimeSlot, DataTimePoint, TimeSeries
from ..models.reconstructors import PeriodicAverageReconstructor, ProphetReconstructor, LinearInterpolationReconstructor
from ..storages import CSVFileStorage

# Setup logging
from .. import logger
logger.setup()

# Test data and temp path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'
TEMP_MODELS_DIR = tempfile.TemporaryDirectory().name

# Ensure reproducibility
import random
import numpy as np
random.seed(0)
np.random.seed(0)


class TestReconstructors(unittest.TestCase):


    def test_BaseReconstructor(self):

        # This test is done using the Linear Interpolator Reconstructor

        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':1}, data_loss=0))
        timeseries.append(DataTimePoint(t=2,  data={'a':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=3,  data={'a':3}, data_loss=1))
        timeseries.append(DataTimePoint(t=4,  data={'a':4}, data_loss=0))

        reconstructor = LinearInterpolationReconstructor()

        # Test the main predict
        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=1)
        self.assertEqual(predicted_data, [{'a': 2.0}])

        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=2)
        self.assertEqual(predicted_data, [{'a': 2.0}, {'a': 3.0}])

        # Not enough widow data for the predict, will raise
        with self.assertRaises(ValueError):
            reconstructor.predict(timeseries, from_i=0,to_i=1)
        with self.assertRaises(ValueError):
            reconstructor.predict(timeseries, from_i=3,to_i=4)

        # Test the apply now
        reconstructed_timeseries = reconstructor.apply(timeseries)
        self.assertEqual(reconstructed_timeseries[1].data['a'],2)

        # Not enough widow data for the apply at the beginning, will [shrink?]
        #timeseries = TimeSeries()
        #timeseries.append(DataTimePoint(t=1,  data={'a':-99}, data_loss=1))
        #timeseries.append(DataTimePoint(t=2,  data={'a':2}, data_loss=0))
        #reconstructor.apply(timeseries)


        # Not enough widow data for the apply at the end, will [shrink?]
        #timeseries = TimeSeries()
        #timeseries.append(DataTimePoint(t=1,  data={'a':-99}, data_loss=1))
        #timeseries.append(DataTimePoint(t=2,  data={'a':2}, data_loss=0))

        # Force the forecaster window to "2" and check that if gaps are separated
        # by a single point, all the are gets marked as to be reconstructed. What we want..?


    def test_PeriodicAverageReconstructor(self):

        # Get test data
        with open(TEST_DATA_PATH + '/csv/temp_slots_1h.csv') as f:
            data=f.read()

        timeseries = TimeSeries()
        for line in data.split('\n'):
            if line:
                start_t = float(line.split(',')[0])
                start_point = TimePoint(t=start_t)
                end_point = TimePoint(t=start_t+3600)
                data_loss = 1-float(line.split(',')[2]) # from coverage to data loss
                value =  float(line.split(',')[1])
                timeseries.append(DataTimeSlot(start=start_point, end=end_point, data={'temperature':value}, data_loss=data_loss))

        # Fit
        reconstructor = PeriodicAverageReconstructor()
        reconstructor.fit(timeseries)

        # Test the evaluate
        evaluation = reconstructor.evaluate(timeseries, limit=100, details=True)
        self.assertAlmostEqual(evaluation['RMSE_1_steps'], 0.131, places=2)
        self.assertAlmostEqual(evaluation['MAE_1_steps'], 0.094, places=2)
        self.assertAlmostEqual(evaluation['RMSE_24_steps'], 0.641, places=2)
        self.assertAlmostEqual(evaluation['MAE_24_steps'], 0.526, places=2)
        self.assertAlmostEqual(evaluation['RMSE'], 0.386, places=2)
        self.assertAlmostEqual(evaluation['MAE'], 0.310, places=2)

        # Test the evaluate on specific steps
        evaluation = reconstructor.evaluate(timeseries, steps=[1,3], limit=100, details=True)
        self.assertAlmostEqual(evaluation['RMSE_1_steps'], 0.131, places=2)
        self.assertAlmostEqual(evaluation['MAE_1_steps'], 0.094, places=2)
        self.assertAlmostEqual(evaluation['RMSE_3_steps'], 0.240, places=2)
        self.assertAlmostEqual(evaluation['MAE_3_steps'], 0.192, places=2)
        self.assertAlmostEqual(evaluation['RMSE'], 0.185, places=2)
        self.assertAlmostEqual(evaluation['MAE'], 0.144, places=2)

        # Test the cross validation
        reconstructor = PeriodicAverageReconstructor()
        cross_validation_results = reconstructor.cross_validate(timeseries, evaluate_steps=[1,3], fit_periodicity=24, evaluate_limit=100, evaluate_details=True)
        self.assertAlmostEqual(cross_validation_results['MAE_3_steps_avg'],  0.2293, places=2)
        self.assertAlmostEqual(cross_validation_results['MAE_3_steps_stdev'], 0.0313, places=2)

        # Test on Points as well
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200)
        reconstructor = PeriodicAverageReconstructor()
        with self.assertRaises(Exception):
            reconstructor.fit(timeseries)

        timeseries = timeseries.resample(600)
        reconstructor.fit(timeseries)

        # TODO: do some actual testing.. not only that "it works"
        _  = reconstructor.apply(timeseries)
        reconstructor.evaluate(timeseries)


    def test_ProphetReconstructor(self):
        try:
            import prophet
        except ImportError:
            print('Skipping Prophet tests as no prophet module installed')
            return

        # Get test data
        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get(limit=200).aggregate(3600)

        # Instantiate
        reconstructor = ProphetReconstructor()

        # Fit
        reconstructor.fit(timeseries)

        # Predict
        predicted_data = reconstructor.predict(timeseries, from_i=2, to_i=5)
        self.assertAlmostEqual(predicted_data['temperature_avg'][0], 23.449303446403178)
        self.assertAlmostEqual(predicted_data['temperature_avg'][1], 23.45425065662797)
        self.assertAlmostEqual(predicted_data['temperature_avg'][2], 23.460265626928496)
        self.assertAlmostEqual(predicted_data['temperature_avg'][3], 23.46643974807204)

        # Predict on first and last as no window
        reconstructor.predict(timeseries, from_i=0, to_i=0)
        reconstructor.predict(timeseries, from_i=len(timeseries)-1, to_i=len(timeseries)-1)


    def test_LinearInterpolationReconstructor(self):

        # Test multivariate
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=1,  data={'a':1, 'b':4}, data_loss=0))
        timeseries.append(DataTimePoint(t=2,  data={'a':-99, 'b':-99}, data_loss=1))
        timeseries.append(DataTimePoint(t=3,  data={'a':3, 'b':2}, data_loss=1))
        timeseries.append(DataTimePoint(t=4,  data={'a':4, 'b':1}, data_loss=0))

        reconstructor = LinearInterpolationReconstructor()

        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=1)
        self.assertEqual(predicted_data, [{'a': 2.0, 'b': 3.0}])


        predicted_data = reconstructor.predict(timeseries, from_i=1,to_i=2)
        self.assertEqual(predicted_data, [{'a': 2.0, 'b': 3.0}, {'a': 3.0, 'b': 2.0}])

        # TODO: move this test in the basic reconstructors
        reconstructed_timeseries = reconstructor.apply(timeseries)
        self.assertEqual(reconstructed_timeseries[0].data, {'a': 1, 'b': 4})
        self.assertEqual(reconstructed_timeseries[1].data, {'a': 2.0, 'b': 3.0})
        self.assertEqual(reconstructed_timeseries[2].data, {'a': 3.0, 'b': 2.0})
        self.assertEqual(reconstructed_timeseries[3].data, {'a': 4, 'b': 1})
        self.assertEqual(reconstructed_timeseries[0].data_indexes, {'data_loss': 0, 'data_reconstructed': 0})
        self.assertEqual(reconstructed_timeseries[1].data_indexes, {'data_loss': 1, 'data_reconstructed': 1})
        self.assertEqual(reconstructed_timeseries[2].data_indexes, {'data_loss': 1, 'data_reconstructed': 1})
        self.assertEqual(reconstructed_timeseries[3].data_indexes, {'data_loss': 0, 'data_reconstructed': 0})

        # Test series
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t=-3,  data={'value':3}))
        timeseries.append(DataTimePoint(t=-2,  data={'value':2}))
        timeseries.append(DataTimePoint(t=-1,  data={'value':1}))
        timeseries.append(DataTimePoint(t=0,  data={'value':0}))
        timeseries.append(DataTimePoint(t=1,  data={'value':1}))
        timeseries.append(DataTimePoint(t=2,  data={'value':2}))
        timeseries.append(DataTimePoint(t=3,  data={'value':3}))
        timeseries.append(DataTimePoint(t=4,  data={'value':4}))
        timeseries.append(DataTimePoint(t=5,  data={'value':5}))
        timeseries.append(DataTimePoint(t=6,  data={'value':6}))
        timeseries.append(DataTimePoint(t=7,  data={'value':7}))
        timeseries.append(DataTimePoint(t=16, data={'value':16}))
        timeseries.append(DataTimePoint(t=17, data={'value':17}))
        timeseries.append(DataTimePoint(t=18, data={'value':18}))

        # Resample for 3 seconds
        resampled_timeseries = timeseries.resample(3)
        # DataTimePoint @ 0.0 (1970-01-01 00:00:00+00:00) with data "{'value': 0.6666666666666666}" and data_loss="0.0"
        # DataTimePoint @ 3.0 (1970-01-01 00:00:03+00:00) with data "{'value': 3.0}" and data_loss="0.0"
        # DataTimePoint @ 6.0 (1970-01-01 00:00:06+00:00) with data "{'value': 6.0}" and data_loss="0.0"
        # DataTimePoint @ 9.0 (1970-01-01 00:00:09+00:00) with data "{'value': 9.0}" and data_loss="1.0"
        # DataTimePoint @ 12.0 (1970-01-01 00:00:12+00:00) with data "{'value': 12.0}" and data_loss="1.0"
        # DataTimePoint @ 15.0 (1970-01-01 00:00:15+00:00) with data "{'value': 15.0}" and data_loss="0.6666666666666667"



        # Fake data to check the recostructor is working correctly
        resampled_timeseries[3].data['value']=-9
        resampled_timeseries[4].data['value']=-12

        # Instantiate the reconstructor
        reconstructor = LinearInterpolationReconstructor()

        # Check parametric param just in case
        self.assertFalse(reconstructor._is_parametric())

        # Apply the reconstructor
        reconstructed_timeseries = reconstructor.apply(resampled_timeseries)

        # Check len
        self.assertEqual(len(reconstructed_timeseries), 6)

        # Check for t = 0
        self.assertEqual(reconstructed_timeseries[0].t, 0)
        self.assertAlmostEqual(reconstructed_timeseries[0].data['value'], 0.6666666666)  # Expected: 0.66..
        self.assertEqual(reconstructed_timeseries[0].data_loss, 0)

        # Check for t = 3
        self.assertEqual(reconstructed_timeseries[1].t, 3)
        self.assertEqual(reconstructed_timeseries[1].data['value'], 3)  # Expected: 3
        self.assertEqual(reconstructed_timeseries[1].data_loss, 0)

        # Check for t = 6
        self.assertEqual(reconstructed_timeseries[2].t, 6)
        self.assertEqual(reconstructed_timeseries[2].data['value'], 6)  # Expected: 6
        self.assertEqual(reconstructed_timeseries[2].data_loss, 0)

        # Check for t = 9
        self.assertEqual(reconstructed_timeseries[3].t, 9)
        self.assertEqual(reconstructed_timeseries[3].data['value'], 9)  # Expected: 9 (fully reconstructed)
        self.assertEqual(reconstructed_timeseries[3].data_loss, 1)

        # Check for t = 12
        self.assertEqual(reconstructed_timeseries[4].t, 12)
        self.assertEqual(reconstructed_timeseries[4].data['value'], 12)  # Expected: 12 (fully reconstructed)
        self.assertEqual(reconstructed_timeseries[4].data_loss, 1)

        # Check for t = 15
        self.assertEqual(reconstructed_timeseries[5].t, 15)
        self.assertEqual(reconstructed_timeseries[5].data['value'], 15)  # Expected: 15
        self.assertAlmostEqual(reconstructed_timeseries[5].data_loss, 0.6666666666)

